from tensorflow.keras.layers import Layer
import tensorflow as tf


class SparseDense(Layer):
    """
    Sparse implementation of the Dense layer. If used in a model, must be saved and loaded via pickle
    """

    def __init__(
        self,
        weight,
        bias,
        activation=None,
        **kwargs
    ):
        """
        Parameters
        ----------
        weight : tf.Tensor
            The kernel tensor
        bias : tf.Tensor
            The bias tensor
        activation : None, str, or keras activation function (default None)
            The activation function to use
        """
        super().__init__(**kwargs)
        self.w = tf.sparse.from_dense(weight)
        self.b = tf.sparse.from_dense(bias)
        self.activation = tf.keras.activations.get(activation)

    def build(self, input_shape):
        """
        Build the layer in preparation to be trained or called. Should not be called directly,
        but rather is called when the layer is added to a model
        """
        pass

    def call(self, inputs):
        """
        This is where the layer's logic lives and is called upon inputs

        Parameters
        ----------
        inputs : TensorFlow Tensor or Tensor-like
            The inputs to the layer

        Returns
        -------
        outputs : TensorFlow Tensor
            The outputs of the layer's logic
        """
        return self.activation(
            tf.sparse.sparse_dense_matmul(
                inputs,
                self.w
            ) + tf.sparse.to_dense(self.b)
        )

    def get_config(self):
        config = super().get_config().copy()
        config['activation'] = tf.keras.activations.serialize(self.activation)
        return config

    @classmethod
    def from_layer(cls, layer):
        """
        Create a layer from an instance of another layer
        """
        weights = layer.get_weights()
        w = weights[0]
        b = weights[1]
        activation = layer.activation
        return cls(
            w,
            b,
            activation
        )

    @classmethod
    def from_config(cls, config):
        return cls(**config)



import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import Layer


class MaskedConv3D(Layer):
    """
    Masked 3-dimensional convolutional layer. For full documentation of the
    convolutional architecture, see the TensorFlow Keras Convolutional3D layer documentation.

    This layer implements masking consistent with the BeyondML API to support
    developing sparse models
    """

    def __init__(
        self,
        filters,
        kernel_size=3,
        padding='same',
        strides=1,
        activation=None,
        use_bias=True,
        kernel_initializer='random_normal',
        bias_initializer='zeros',
        mask_initializer='ones',
        **kwargs
    ):
        """
        Parameters
        ----------
        filters : int
            The number of convolutional filters to apply
        kernel_size : int or tuple of ints (default 3)
            The kernel size in height, width, and depth
        padding : str (default 'same')
            Either 'same' or 'valid', the padding to use during convolution
        strides : int or tuple of ints (default 1)
            Stride lengths in each dimension to use during convolution
        activation : None, str, or function (default None)
            Activation function to use on the outputs
        use_bias : bool (default True)
            Whether to use a bias calculation on the outputs
        kernel_initializer : str or keras initialization function (default 'random_normal')
            The weight initialization function to use
        bias_initializer : str or keras initialization function (default 'zeros')
            The bias initialization function to use
        mask_initializer : str or keras initialization function (default 'ones')
            The mask initialization function to use
        """

        super().__init__(**kwargs)
        self.filters = int(filters) if not isinstance(
            filters, int) else filters
        self.kernel_size = kernel_size
        self.padding = padding
        self.strides = tuple(strides) if isinstance(strides, list) else strides
        self.activation = tf.keras.activations.get(activation)
        self.use_bias = use_bias
        self.kernel_initializer = tf.keras.initializers.get(kernel_initializer)
        self.bias_initializer = tf.keras.initializers.get(bias_initializer)
        self.mask_initializer = tf.keras.initializers.get(mask_initializer)

    @property
    def kernel_size(self):
        return self._kernel_size

    @kernel_size.setter
    def kernel_size(self, value):
        if isinstance(value, int):
            self._kernel_size = (value, value, value)
        else:
            self._kernel_size = value

    def build(self, input_shape):
        """
        Build the layer in preparation to be trained or called. Should not be called directly,
        but rather is called when the layer is added to a model
        """
        self.w = self.add_weight(
            shape=(self.kernel_size[0], self.kernel_size[1],
                   self.kernel_size[2], input_shape[-1], self.filters),
            initializer=self.kernel_initializer,
            trainable=True,
            name='weights'
        )
        self.w_mask = self.add_weight(
            shape=self.w.shape,
            initializer=self.mask_initializer,
            trainable=False,
            name='weights_mask'
        )

        if self.use_bias:
            self.b = self.add_weight(
                shape=(self.filters,),
                initializer=self.bias_initializer,
                trainable=True,
                name='bias'
            )
            self.b_mask = self.add_weight(
                shape=self.b.shape,
                initializer=self.mask_initializer,
                trainable=False,
                name='bias_mask'
            )

    def call(self, inputs):
        """
        This is where the layer's logic lives and is called upon inputs

        Parameters
        ----------
        inputs : TensorFlow Tensor or Tensor-like
            The inputs to the layer

        Returns
        -------
        outputs : TensorFlow Tensor
            The outputs of the layer's logic
        """
        conv_output = tf.nn.convolution(
            inputs,
            self.w * self.w_mask,
            padding=self.padding.upper() if isinstance(
                self.padding, str) else self.padding,
            strides=self.strides,
            data_format='NDHWC'
        )
        if self.use_bias:
            conv_output = conv_output + (self.b * self.b_mask)
        return self.activation(conv_output)

    def get_config(self):
        config = super().get_config().copy()
        config.update(
            {
                'filters': self.filters,
                'kernel_size': list(self.kernel_size),
                'padding': self.padding,
                'strides': self.strides,
                'activation': tf.keras.activations.serialize(self.activation),
                'use_bias': self.use_bias,
                'kernel_initializer': tf.keras.initializers.serialize(self.kernel_initializer),
                'bias_initializer': tf.keras.initializers.serialize(self.bias_initializer),
                'mask_initializer': tf.keras.initializers.serialize(self.mask_initializer)
            }
        )
        return config

    def set_masks(self, new_masks):
        """
        Set the masks for the layer

        Parameters
        ----------
        new_masks : list of arrays or array-likes
            The new masks to set for the layer
        """

        if not self.use_bias:
            self.set_weights(
                [self.w.numpy() * new_masks[0].astype(np.float32),
                 new_masks[0].astype(np.float32)]
            )
        else:
            self.set_weights(
                [self.w.numpy() * new_masks[0].astype(np.float32), self.b.numpy() * new_masks[1].astype(
                    np.float32), new_masks[0].astype(np.float32), new_masks[1].astype(np.float32)]
            )

    @classmethod
    def from_config(cls, config):
        return cls(
            filters=config['filters'],
            kernel_size=config['kernel_size'],
            padding=config['padding'],
            strides=config['strides'],
            activation=config['activation'],
            use_bias=config['use_bias'],
            kernel_initializer=config['kernel_initializer'],
            bias_initializer=config['bias_initializer'],
            mask_initializer=config['mask_initializer']
        )



import tensorflow as tf
from beyondml.tflow.layers import MultiDense, SelectorLayer


def build_transformer_block(
        input_shape,
        embed_dim,
        num_heads,
        neurons,
        dropout_rate=0.1,
):
    """
    Build a Transformer Block

    Parameters
    ----------
    input_shape : int or tuple of int
        The input shape for the model to use
    embed_dim : int
        The dimension of the embedding
    num_heads : int
        The number of attention heads to use
    neurons : int
        The number of hidden neurons to use in the hidden layer
    dropout_rate : float (default 0.1)
        Rate at which dropout is applied
    value_dim : int or None (default None)
        The dimension to use for the `value` matrix, if provided

    Returns
    -------
    transformer_block : TensorFlow keras Functional model
        The transformer block, which can then be used alone or as
        a layer in another model
    """
    input_layer = tf.keras.layers.Input(input_shape)
    query = MultiDense(embed_dim)([input_layer] * num_heads)
    key = MultiDense(embed_dim)([input_layer] * num_heads)
    value = MultiDense(embed_dim)([input_layer] * num_heads)

    query_selectors = [
        SelectorLayer(i)(query) for i in range(num_heads)
    ]
    key_selectors = [
        SelectorLayer(i)(key) for i in range(num_heads)
    ]
    value_selectors = [
        SelectorLayer(i)(value) for i in range(num_heads)
    ]
    attention_layers = [
        tf.keras.layers.Attention()([query_selectors[i], key_selectors[i], value_selectors[i]]) for i in range(num_heads)
    ]
    concat = tf.keras.layers.Concatenate()(attention_layers)
    merge = tf.keras.layers.Reshape((input_shape[0], -1))(concat)

    x = tf.keras.layers.Dropout(dropout_rate)(merge)
    out1 = tf.keras.layers.LayerNormalization(epsilon=1e-6)(x)
    x = tf.keras.layers.Dense(neurons, activation='relu')(out1)
    x = tf.keras.layers.Dense(embed_dim * num_heads)(x)
    x = tf.keras.layers.Dropout(dropout_rate)(x)
    x = tf.keras.layers.Add()([out1, x])
    output_layer = tf.keras.layers.LayerNormalization(epsilon=1e-6)(x)

    return tf.keras.models.Model(input_layer, output_layer)


def build_token_position_embedding_block(
    sequence_length,
    vocab_size,
    embed_dim
):
    """
    Builds a token and position embedding block

    Parameters
    ----------
    sequence_length : int
        The length of each sequence
    vocab_size : int
        The size of the vocabulary used
    embed_dim : int
        The desired embedding dimension

    Returns
    -------
    embedding_block : TensorFlow keras Functional model
        The embedding block, which can be used alone or
        as a layer in another model
    """
    tok_input = tf.keras.layers.Input(sequence_length)
    pos_input = tf.keras.layers.Input(sequence_length)

    tok_embed = tf.keras.layers.Embedding(
        vocab_size, output_dim=embed_dim)(tok_input)
    pos_embed = tf.keras.layers.Embedding(
        sequence_length, output_dim=embed_dim)(pos_input)
    output_layer = tf.keras.layers.Add()([tok_embed, pos_embed])

    return tf.keras.models.Model([tok_input, pos_input], output_layer)



import torch


class FilterLayer(torch.nn.Module):
    """
    Layer which filters input data, either returning values or all zeros depending on state
    """

    def __init__(
        self,
        is_on=True,
        device=None,
        dtype=None
    ):
        """
        Parameters
        ----------
        is_on : bool (default False)
            Whether the layer is on or off
        """

        super().__init__()
        self.is_on = is_on
        self.factory_kwargs = {'device': device, 'dtype': dtype}

    @property
    def is_on(self):
        return self._is_on

    @is_on.setter
    def is_on(self, value):
        if not isinstance(value, bool):
            raise TypeError('is_on must be Boolean')
        self._is_on = value

    def forward(self, inputs):
        """
        Call the layer on input data

        Parameters
        ----------
        inputs : torch.Tensor
            Inputs to call the layer's logic on

        Returns
        -------
        results : torch.Tensor
            The results of the layer's logic
        """
        if self.is_on:
            return inputs
        else:
            return torch.zeros_like(inputs, **self.factory_kwargs)

    def turn_on(self):
        """
        Turn on the layer
        """
        self.is_on = True

    def turn_off(self):
        """
        Turn off the layer
        """
        self.is_on = False



import numpy as np
import torch


class MultiMaskedConv2D(torch.nn.Module):
    """
    Multi 2D Convolutional layer which supports masking and pruning
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        num_tasks,
        kernel_size=3,
        padding='same',
        strides=1,
        device=None,
        dtype=None
    ):
        """
        Parameters
        ----------
        in_channels : int
            The number of input channels for all inputs
        out_channels : int
            The number of filters to use for all inputs
        num_tasks : int
            The number of tasks to initialize the layer for
        kernel_size : int (default 3)
            The kernel size to use
        padding : str or int (default 'same')
            The padding to use
        strides : int or tuple (default 1)
            The strides to use
        """

        super().__init__()
        factory_kwargs = {'device': device, 'dtype': dtype}
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.num_tasks = num_tasks
        self.kernel_size = kernel_size
        self.padding = padding
        self.strides = strides

        filters = torch.Tensor(
            self.num_tasks,
            self.out_channels,
            self.in_channels,
            self.kernel_size[0],
            self.kernel_size[1],
        ).to(**factory_kwargs)
        filters = torch.nn.init.kaiming_normal_(filters, a=np.sqrt(5))
        self.w = torch.nn.Parameter(filters)
        self.register_buffer(
            'w_mask', torch.ones_like(self.w, **factory_kwargs))

        bias = torch.zeros(self.num_tasks, out_channels, **factory_kwargs)
        self.b = torch.nn.Parameter(bias)
        self.register_buffer(
            'b_mask', torch.ones_like(self.b, **factory_kwargs))

    @property
    def in_channels(self):
        return self._in_channels

    @in_channels.setter
    def in_channels(self, value):
        if not isinstance(value, int):
            raise TypeError('in_channels must be int')
        self._in_channels = value

    @property
    def out_channels(self):
        return self._out_channels

    @out_channels.setter
    def out_channels(self, value):
        if not isinstance(value, int):
            raise TypeError('out_channels must be int')
        self._out_channels = value

    @property
    def kernel_size(self):
        return self._kernel_size

    @kernel_size.setter
    def kernel_size(self, value):
        if isinstance(value, int):
            value = (value, value)
        elif isinstance(value, tuple):
            if not all([isinstance(val, int) for val in value]) and len(value) == 2:
                raise ValueError('If tuple, kernel_size must be two integers')
        else:
            raise TypeError('kernel_size must be int or tuple')
        self._kernel_size = value

    def forward(self, inputs):
        """
        Call the layer on input data

        Parameters
        ----------
        inputs : torch.Tensor
            Inputs to call the layer's logic on

        Returns
        -------
        results : torch.Tensor
            The results of the layer's logic
        """
        outputs = []
        for i in range(len(inputs)):
            outputs.append(
                torch.nn.functional.conv2d(
                    inputs[i],
                    self.w[i] * self.w_mask[i],
                    self.b[i] * self.b_mask[i],
                    stride=self.strides,
                    padding=self.padding
                )
            )
        return outputs

    def prune(self, percentile):
        """
        Prune the layer by updating the layer's mask

        Parameters
        ----------
        percentile : int
            Integer between 0 and 99 which represents the proportion of weights to be inactive

        Notes
        -----
        Acts on the layer in place
        """
        w_copy = np.abs(self.w.detach().cpu().numpy())
        b_copy = np.abs(self.b.detach().cpu().numpy())
        new_w_mask = np.zeros_like(w_copy)
        new_b_mask = np.zeros_like(b_copy)

        for task_num in range(self.num_tasks):
            if task_num != 0:
                for prev_idx in range(task_num):
                    w_copy[task_num][new_w_mask[prev_idx] == 1] = 0
                    b_copy[task_num][new_b_mask[prev_idx] == 1] = 0

            w_percentile = np.percentile(w_copy[task_num], percentile)
            b_percentile = np.percentile(b_copy[task_num], percentile)

            new_w_mask[task_num] = (
                w_copy[task_num] >= w_percentile).astype(int)
            new_b_mask[task_num] = (
                b_copy[task_num] >= b_percentile).astype(int)

        self.w_mask[:] = torch.Tensor(new_w_mask)
        self.b_mask[:] = torch.Tensor(new_b_mask)

        self.w = torch.nn.Parameter(
            self.w * self.w_mask
        )
        self.b = torch.nn.Parameter(
            self.b * self.b_mask
        )



import torch


class MultiDense(torch.nn.Module):
    """
    Multi-Fully-Connected layer initialized with weights rather than hyperparameters
    """

    def __init__(
            self,
            weight,
            bias,
            device=None,
            dtype=None
    ):
        """
        Parameters
        ----------
        weight : torch.Tensor or Tensor-like
            The weight tensor to use
        bias : torch.Tensor or Tensor-like
            The bias tensor to use
        """

        factory_kwargs = {'device': device, 'dtype': dtype}
        super().__init__()
        self.w = torch.nn.Parameter(torch.Tensor(weight).to(**factory_kwargs))
        self.b = torch.nn.Parameter(torch.Tensor(bias).to(**factory_kwargs))

    def forward(self, inputs):
        """
        Call the layer on input data

        Parameters
        ----------
        inputs : torch.Tensor
            Inputs to call the layer's logic on

        Returns
        -------
        results : torch.Tensor
            The results of the layer's logic
        """
        outputs = []
        for i in range(len(inputs)):
            out = torch.mm(inputs[i], self.w[i])
            out = torch.add(out, self.b[i])
            outputs.append(out)
        return outputs



import torch


class Dense(torch.nn.Module):
    """
    Fully-connected layer initialized directly with weights, rather than hyperparameters
    """

    def __init__(
        self,
        weight,
        bias,
        device=None,
        dtype=None
    ):
        """
        Parameters
        ----------
        weight : torch.Tensor or Tensor-like
            The weight matrix to use
        bias : torch.Tensor or Tensor-like
            The bias vector to use
        """
        factory_kwargs = {'device': device, 'dtype': dtype}
        super().__init__()
        self.w = torch.nn.Parameter(torch.Tensor(weight).to(**factory_kwargs))
        self.b = torch.nn.Parameter(torch.Tensor(bias).to(**factory_kwargs))

    def forward(self, inputs):
        """
        Call the layer on input data

        Parameters
        ----------
        inputs : torch.Tensor
            Inputs to call the layer's logic on

        Returns
        -------
        results : torch.Tensor
            The results of the layer's logic
        """
        out = torch.mm(inputs, self.w)
        out = torch.add(out, self.b)
        return out



from torch.nn import Module


class SelectorLayer(Module):
    """
    Layer which selects an individual input based on index and only returns that one
    """

    def __init__(
        self,
        sel_index
    ):
        """
        Parameters
        ----------
        sel_index : int
            The index of inputs to select
        """
        super().__init__()
        self.sel_index = sel_index

    @property
    def sel_index(self):
        return self._sel_index

    @sel_index.setter
    def sel_index(self, value):
        if not isinstance(value, int):
            raise TypeError('sel_index must be integer-valued')
        self._sel_index = value

    def forward(self, inputs):
        """
        Call the layer on input data

        Parameters
        ----------
        inputs : torch.Tensor
            Inputs to call the layer's logic on

        Returns
        -------
        results : torch.Tensor
            The results of the layer's logic
        """
        return inputs[self.sel_index]



import torch


class MultiMaxPool3D(torch.nn.Module):
    """
    Multitask implementation of 2-dimensional Max Pooling layer
    """

    def __init__(
        self,
        kernel_size,
        stride=None,
        padding=0,
        dilation=1
    ):
        """
        Parameters
        ----------
        kernel_size : int or tuple
            The kernel size to use
        stride : int, tuple, or None (default None)
            The stride to use.  If None, defaults to kernel_size
        padding : int (default 0)
            The padding to use
        dilation : int (default 1)
            The dilation to use
        """
        self.kernel_size = kernel_size
        self.stride = stride if stride else self.kernel_size
        self.padding = padding
        self.dilation = dilation

    def forward(
        self,
        inputs
    ):
        """
        Call the layer on input data

        Parameters
        ----------
        inputs : torch.Tensor
            Inputs to call the layer's logic on

        Returns
        -------
        results : torch.Tensor
            The results of the layer's logic
        """
        outputs = []
        for i in range(len(inputs)):
            outputs.append(
                torch.nn.functional.max_pool3d(
                    input=inputs[i],
                    kernel_size=self.kernel_size,
                    stride=self.stride,
                    padding=self.padding,
                    dilation=self.dilation
                )
            )
        return outputs



import numpy as np
import torch


class MaskedDense(torch.nn.Module):
    """
    Masked fully-connected layer
    """

    def __init__(
        self,
        in_features,
        out_features,
        device=None,
        dtype=None
    ):
        """
        Parameters
        ----------
        in_features : int
            The number of features input to the layer
        out_features : int
            The number of features to be output by the layer.
            Also considered the number of artificial neurons
        """

        super().__init__()
        factory_kwargs = {'device': device, 'dtype': dtype}
        self.in_features = in_features
        self.out_features = out_features

        weight = torch.Tensor(
            in_features,
            out_features,
        ).to(**factory_kwargs)
        weight = torch.nn.init.kaiming_normal_(weight, a=np.sqrt(5))
        self.w = torch.nn.Parameter(weight)
        self.register_buffer(
            'w_mask', torch.ones_like(self.w, **factory_kwargs))

        bias = torch.zeros(out_features, **factory_kwargs)
        self.b = torch.nn.Parameter(bias)
        self.register_buffer('b_mask', torch.ones_like(bias, **factory_kwargs))

    def forward(self, inputs):
        """
        Call the layer on input data

        Parameters
        ----------
        inputs : torch.Tensor
            Inputs to call the layer's logic on

        Returns
        -------
        results : torch.Tensor
            The results of the layer's logic
        """
        weight = self.w * self.w_mask
        bias = self.b * self.b_mask
        out = torch.mm(inputs, weight)
        out = torch.add(out, bias)
        return out

    def prune(self, percentile):
        """
        Prune the layer by updating the layer's mask

        Parameters
        ----------
        percentile : int
            Integer between 0 and 99 which represents the proportion of weights to be inactive

        Notes
        -----
        Acts on the layer in place
        """
        w_copy = np.abs(self.w.detach().cpu().numpy())
        b_copy = np.abs(self.b.detach().cpu().numpy())
        w_percentile = np.percentile(w_copy, percentile)
        b_percentile = np.percentile(b_copy, percentile)

        new_w_mask = torch.Tensor(
            (w_copy >= w_percentile).astype(int))
        new_b_mask = torch.Tensor(
            (b_copy >= b_percentile).astype(int))
        self.w_mask[:] = new_w_mask
        self.b_mask[:] = new_b_mask

        self.w = torch.nn.Parameter(
            self.w.detach() * self.w_mask
        )
        self.b = torch.nn.Parameter(
            self.b.detach() * self.b_mask
        )



"""Custom layers to use when building MANN models"""

from .MultitaskNormalization import MultitaskNormalization
from .FilterLayer import FilterLayer
from .SumLayer import SumLayer
from .MaskedDense import MaskedDense
from .MaskedConv2D import MaskedConv2D
from .MaskedConv3D import MaskedConv3D
from .SelectorLayer import SelectorLayer
from .MultiMaskedDense import MultiMaskedDense
from .MultiMaskedConv2D import MultiMaskedConv2D
from .MultiMaskedConv3D import MultiMaskedConv3D
from .MultiDense import MultiDense
from .MultiConv2D import MultiConv2D
from .MultiConv3D import MultiConv3D
from .MultiMaxPool2D import MultiMaxPool2D
from .MultiMaxPool3D import MultiMaxPool3D
from .SparseDense import SparseDense
from .SparseConv2D import SparseConv2D
from .SparseConv3D import SparseConv3D
from .SparseMultiDense import SparseMultiDense
from .SparseMultiConv2D import SparseMultiConv2D
from .SparseMultiConv3D import SparseMultiConv3D



"""
BeyondML (formerly MANN) is a Python package which enables creating sparse multitask artificial neural networks (MANNs)
compatible with [TensorFlow](https://tensorflow.org) and [PyTorch](https://pytorch.org). This package
contains custom layers and utilities to facilitate the training and optimization of models using the
Reduction of Sub-Network Neuroplasticity (RSN2) training procedure developed by [AI Squared, Inc](https://squared.ai).

### Installation

This package is available through [PyPi](https://pypi.org) and can be installed via the following command:

```bash
pip install beyondml
```

### Capabilities

There are two major subpackages within the BeyondML package, the `beyondml.tflow` and the `beyondml.pt` packages.
The `beyondml.tflow` package contains functionality for building multitask models using TensorFlow, and the
`beyondml.pt` package contains functionality for building multitask models using PyTorch.
"""

__version__ = '0.1.3.dev2'
__dev__ = True

import beyondml.tflow
import beyondml.pt



import numpy as np
import torch


class MaskedConv2D(torch.nn.Module):
    """
    Masked 2D Convolutional layer
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=3,
        padding='same',
        strides=1,
        device=None,
        dtype=None
    ):
        """
        Parameters
        ----------
        in_channels : int
            The number of channels for input data
        out_channels : int
            The number of filters to use
        kernel_size : int or tuple (default 3)
            The kernel size to use
        padding : str or int (default 'same')
            Padding to use
        strides : int or tuple (default 1)
            The number of strides to use
        """

        super().__init__()
        factory_kwargs = {'device': device, 'dtype': dtype}
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.padding = padding
        self.strides = strides

        filters = torch.Tensor(
            self.out_channels,
            self.in_channels,
            self.kernel_size[0],
            self.kernel_size[1],
        ).to(**factory_kwargs)
        filters = torch.nn.init.kaiming_normal_(filters, a=np.sqrt(5))
        self.w = torch.nn.Parameter(filters)
        self.register_buffer(
            'w_mask', torch.ones_like(self.w, **factory_kwargs))

        bias = torch.zeros(out_channels, **factory_kwargs)
        self.b = torch.nn.Parameter(bias)
        self.register_buffer(
            'b_mask', torch.ones_like(self.b, **factory_kwargs))

    @property
    def in_channels(self):
        return self._in_channels

    @in_channels.setter
    def in_channels(self, value):
        if not isinstance(value, int):
            raise TypeError('in_channels must be int')
        self._in_channels = value

    @property
    def out_channels(self):
        return self._out_channels

    @out_channels.setter
    def out_channels(self, value):
        if not isinstance(value, int):
            raise TypeError('out_channels must be int')
        self._out_channels = value

    @property
    def kernel_size(self):
        return self._kernel_size

    @kernel_size.setter
    def kernel_size(self, value):
        if isinstance(value, int):
            value = (value, value)
        elif isinstance(value, tuple):
            if not all([isinstance(val, int) for val in value]) and len(value) == 2:
                raise ValueError('If tuple, kernel_size must be two integers')
        else:
            raise TypeError('kernel_size must be int or tuple')
        self._kernel_size = value

    def forward(self, inputs):
        """
        Call the layer on input data

        Parameters
        ----------
        inputs : torch.Tensor
            Inputs to call the layer's logic on

        Returns
        -------
        results : torch.Tensor
            The results of the layer's logic
        """
        return torch.nn.functional.conv2d(
            inputs,
            self.w * self.w_mask,
            self.b * self.b_mask,
            stride=self.strides,
            padding=self.padding
        )

    def prune(self, percentile):
        """
        Prune the layer by updating the layer's mask

        Parameters
        ----------
        percentile : int
            Integer between 0 and 99 which represents the proportion of weights to be inactive

        Notes
        -----
        Acts on the layer in place
        """
        w_copy = np.abs(self.w.detach().cpu().numpy())
        b_copy = np.abs(self.b.detach().cpu().numpy())
        w_percentile = np.percentile(w_copy, percentile)
        b_percentile = np.percentile(b_copy, percentile)

        new_w_mask = torch.Tensor(
            (w_copy >= w_percentile).astype(int))
        new_b_mask = torch.Tensor(
            (b_copy >= b_percentile).astype(int))
        self.w_mask[:] = new_w_mask
        self.b_mask[:] = new_b_mask

        self.w = torch.nn.Parameter(self.w.detach() * self.w_mask)
        self.b = torch.nn.Parameter(self.b.detach() * self.b_mask)



import torch


class MultitaskNormalization(torch.nn.Module):
    """
    Layer which normalizes a set of inputs to sum to 1
    """

    def __init__(
        self,
        device=None,
        dtype=None
    ):

        super().__init__()
        self.factory_kwargs = {'device': device, 'dtype': dtype}

    def forward(self, inputs):
        """
        Call the layer on input data

        Parameters
        ----------
        inputs : torch.Tensor or list of Tensors
            Inputs to call the layer's logic on

        Returns
        -------
        results : torch.Tensor or list of Tensors
            The results of the layer's logic
        """
        s = 0
        for i in inputs:
            s += 1
        return [i / s for i in inputs]



import tensorflow as tf
from tensorflow.keras.layers import Layer


class SumLayer(Layer):
    """
    Layer which adds all inputs together. All inputs must have compatible shapes

    Example:

    >>> # Create a model with just a SumLayer and two inputs
    >>> input_1 = tf.keras.layers.Input(10)
    >>> input_2 = tf.keras.layers.Input(10)
    >>> sum_layer = mann.layers.SumLayer()([input_1, input_2])
    >>> model = tf.keras.models.Model([input_1, input_2], sum_layer)
    >>> model.compile()
    >>> # Call the model
    >>> data = np.arange(10).reshape((1, 10))
    >>> model.predict([data, data])
    array([[ 0.,  2.,  4.,  6.,  8., 10., 12., 14., 16., 18.]], dtype=float32)

    """

    def __init__(self, **kwargs):
        super(SumLayer, self).__init__(**kwargs)

    def call(self, inputs):
        """
        This is where the layer's logic lives and is called upon inputs

        Parameters
        ----------
        inputs : TensorFlow Tensor or Tensor-like
            The inputs to the layer

        Returns
        -------
        outputs : TensorFlow Tensor
            The outputs of the layer's logic
        """
        return tf.add_n(inputs)

    def get_config(self):
        return super().get_config().copy()

    @classmethod
    def from_config(cls, config):
        return cls(**config)



from .MaskedDense import MaskedDense
import numpy as np
import torch


class MaskedMultiHeadAttention(torch.nn.Module):
    """
    Masked Multi-Headed Attention Layer
    """

    def __init__(
            self,
            embed_dim,
            num_heads,
            dropout=0,
            batch_first=False,
            device=None,
            dtype=None
    ):
        """
        Parameters
        ----------
        embed_dim : int
            The embedding dimension
        num_heads : int
            The number of attention heads
        dropout : float (default 0)
            The dropout rate to apply
        batch_first : bool (default False)
            Whether the batch dimension is first
        """

        super().__init__()

        factory_kwargs = {'device': device, 'dtype': dtype}
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.dropout = dropout
        self.batch_first = batch_first

        self.head_dim = embed_dim // num_heads

        if self.head_dim * self.num_heads != embed_dim:
            raise ValueError('num_heads must evenly divide embed_dim')

        in_proj_weight = torch.Tensor(
            3 * embed_dim, embed_dim).to(**factory_kwargs)
        in_proj_weight = torch.nn.init.xavier_uniform_(in_proj_weight)
        self.in_proj_weight = torch.nn.Parameter(in_proj_weight)
        self.register_buffer('in_proj_weight_mask', torch.ones_like(
            self.in_proj_weight, **factory_kwargs))

        self.in_proj_bias = torch.nn.Parameter(
            torch.zeros((3 * embed_dim), **factory_kwargs))
        self.register_buffer('in_proj_bias_mask', torch.ones_like(
            self.in_proj_bias, **factory_kwargs))

        self.out_proj = MaskedDense(
            embed_dim, embed_dim, **factory_kwargs)
        self.out_proj_weight = self.out_proj.w
        self.out_proj_weight_mask = self.out_proj.w_mask
        self.out_proj_bias = self.out_proj.b
        self.out_proj_bias_mask = self.out_proj.b_mask

    def forward(
            self,
            query,
            key,
            value,
            key_padding_mask=None,
            need_weights=True,
            attn_mask=None,
            average_attn_weights=True
    ):
        """
        Call the layer on input data

        Parameters
        ----------
        query : torch Tensor
            Query tensor
        key : torch Tensor
            Key tensor
        value : torch Tensor
            Value tensor
        key_padding_mask : None or torch Tensor (default None)
            If specified, a mask indicating which elements in ``key`` to ignore
        need_weights : Bool (default True)
            If specified, returns ``attn_output_weights`` as well as ``attn_outputs``
        attn_mask : None or torch Tensor (default None)
            If specified, a 2D or 3D mask preventing attention
        average_attn_weights : Bool (default True)
            If True, indicates that returned ``attn_weights`` should be averaged across heads
        """

        is_batched = query.dim() == 3

        if self.batch_first and is_batched:
            if key is value:
                if query is key:
                    query = key = value = query.transpose(1, 0)
                else:
                    query, key = [x.transpose(1, 0) for x in (query, key)]
                    value = key
            else:
                query, key, value = [x.transpose(
                    1, 0) for x in (query, key, value)]

        attn_output, attn_output_weights = torch.nn.functional.multi_head_attention_forward(
            query,
            key,
            value,
            self.embed_dim,
            self.num_heads,
            self.in_proj_weight * self.in_proj_weight_mask,
            self.in_proj_bias * self.in_proj_bias_mask,
            training=self.training,
            key_padding_mask=key_padding_mask,
            need_weights=need_weights,
            attn_mask=attn_mask,
            average_attn_weights=average_attn_weights
        )

        if self.batch_first and is_batched:
            return attn_output.transpose(1, 0), attn_output_weights
        else:
            return attn_output, attn_output_weights

    def prune(self, percentile):
        """
        Prune the layer by updating the layer's mask

        Parameters
        ----------
        percentile : int
            Integer between 0 and 99 which represents the proportion of weights to be made inactive

        Notes
        -----
        Acts on the layer in place
        """
        w_copy = np.abs(self.in_proj_weight.detach().cpu().numpy())
        b_copy = np.abs(self.in_proj_bias.detach().cpu().numpy())
        w_percentile = np.percentile(w_copy, percentile)
        b_percentile = np.percentile(b_copy, percentile)

        new_w_mask = torch.Tensor(
            (w_copy >= w_percentile).astype(int))
        new_b_mask = torch.Tensor(
            (b_copy >= b_percentile).astype(int))
        self.in_proj_weight_mask[:] = new_w_mask
        self.in_proj_bias_mask[:] = new_b_mask

        self.in_proj_weight = torch.nn.Parameter(
            self.in_proj_weight * self.in_proj_weight_mask
        )
        self.in_proj_bias = torch.nn.Parameter(
            self.in_proj_bias * self.in_proj_bias_mask
        )
        self.out_proj.prune(percentile)



import torch


class SparseMultiConv3D(torch.nn.Module):
    """
    Sparse implementation of a Multitask 3D Convolutional layer, expected to be converted from a
    trained, pruned layer
    """

    def __init__(
        self,
        kernel,
        bias,
        padding='same',
        strides=1,
        device=None,
        dtype=None
    ):
        """
        Parameters
        ----------
        kernel : torch.Tensor or Tensor-like
            The kernel to use
        bias : torch.Tensor or Tensor-like
            The bias to use
        padding : str or int (default 'same')
            The padding to use
        strides : int or tuple (default 1)
            The strides to use
        """

        factory_kwargs = {'device': device, 'dtype': dtype}
        super().__init__()
        for i in range(kernel.shape[0]):
            self.register_buffer(
                f'w_{i}',
                torch.Tensor(kernel[i]).to(**factory_kwargs).to_sparse()
            )
            self.register_buffer(
                f'b_{i}',
                torch.Tensor(bias[i]).to(**factory_kwargs).to_sparse()
            )

        self.padding = padding
        self.strides = strides

    def forward(
        self,
        inputs
    ):
        """
        Call the layer on input data

        Parameters
        ----------
        inputs : torch.Tensor
            Inputs to call the layer's logic on

        Returns
        -------
        results : torch.Tensor
            The results of the layer's logic
        """

        outputs = []
        for i in range(len(inputs)):
            outputs.append(
                torch.nn.functional.conv3d(
                    inputs[i],
                    self.get_buffer(f'w_{i}').to_dense(),
                    self.get_buffer(f'b_{i}').to_dense(),
                    stride=self.strides,
                    padding=self.padding
                )
            )
        return outputs



"""
## PyTorch compatibility for building MANN models

The `beyondml.pt` subpackage contains layers and utilities for creating and pruning models using [PyTorch](https://pytorch.org).
The package contains two subpackages, the `beyondml.pt.layers` package, and the `beyondml.pt.utils` package.

Within the `layers` package, there is current functionality for the the following layers:
- `beyondml.pt.layers.Conv2D`
- `beyondml.pt.layers.Dense`
- `beyondml.pt.layers.FilterLayer`
- `beyondml.pt.layers.MaskedConv2D`
- `beyondml.pt.layers.MaskedDense`
- `beyondml.pt.layers.MultiConv2D`
- `beyondml.pt.layers.MultiDense`
- `beyondml.pt.layers.MultiMaskedConv2D`
- `beyondml.pt.layers.MultiMaskedDense`
- `beyondml.pt.layers.SelectorLayer`
- `beyondml.pt.layers.SparseConv2D`
- `beyondml.pt.layers.SparseDense`
- `beyondml.pt.layers.SparseMultiConv2D`
- `beyondml.pt.layers.SparseMultiDense`

Within the `beyondml.pt.utils` package, there is currently only one function, the `prune_model` function. Because of
the openness of developing with PyTorch in comparison to TensorFlow, there is far less functionality that
can be supplied directly via BeyondML. Instead, for converting models from training to inference, the user
is left to devise the best way to do so by building his or her own classes.

### Best Practices for Pruning
In order to use the `utils.prune_model` function, the model itself must have a `.layers` property. This property
is used to determine which layers can be pruned. **Only layers which support pruning and which are included in the
`.layers` property are pruned,** meaning the user can determine which exact layers in the model he or she wants
pruned. Alternatively, the user can create their own pruning function or method on the class itself and prune that way,
utilizing each of the `.prune()` methods of the layers provided.
"""

import beyondml.pt.layers
import beyondml.pt.utils



from tensorflow.keras.layers import Layer


class MultitaskNormalization(Layer):
    """
    Multitask layer which normalizes all inputs to sum to 1
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def build(self, input_shape):
        pass

    def call(self, inputs):
        """
        This is where the layer's logic lives and is called upon inputs

        Parameters
        ----------
        inputs : TensorFlow Tensor or Tensor-like
            The inputs to the layer

        Returns
        -------
        outputs : TensorFlow Tensor
            The outputs of the layer's logic
        """
        s = 0
        for i in inputs:
            s += i
        return [i / s for i in inputs]

    def get_config(self):
        return super().get_config()

    @classmethod
    def from_config(cls, config):
        return cls(**config)



from multiprocessing.sharedctypes import Value
import tensorflow as tf
from tensorflow.keras.layers import Layer


class MultiConv3D(Layer):
    """
    Multitask 3-dimensional convolutional layer

    This layer implements multiple stacks of convolutional weights to account for different ways individual
    neurons activate for various tasks. It is expected that to train using the RSN2 algorithm that MultiMaskedConv3D
    layers be used during training and then those layers be converted to this layer type.
    """

    def __init__(
        self,
        filters,
        kernel_size=3,
        padding='same',
        strides=1,
        use_bias=True,
        activation=None,
        kernel_initializer='random_normal',
        bias_initializer='zeros',
        **kwargs
    ):
        """
        Parameters
        ----------
        filters : int
            The number of convolutional filters to apply
        kernel_size : int or tuple of ints (default 3)
            The kernel size in height and width
        padding : str (default 'same')
            Either 'same' or 'valid', the padding to use during convolution
        strides : int or tuple of ints
            Stride lengths to use during convolution
        use_bias : bool (default True)
            Whether to use a bias calculation on the outputs
        activation : None, str, or function (default None)
            Activation function to use on the outputs
        kernel_initializer : str or keras initialization function (default 'random_normal')
            The weight initialization function to use
        bias_initializer : str or keras initialization function (default 'zeros')
            The bias initialization function to use

        """
        super().__init__(**kwargs)
        self.filters = int(filters) if not isinstance(
            filters, int) else filters
        self.kernel_size = kernel_size
        self.padding = padding
        self.strides = tuple(strides) if isinstance(strides, list) else strides
        self.activation = tf.keras.activations.get(activation)
        self.use_bias = use_bias
        self.kernel_initializer = tf.keras.initializers.get(kernel_initializer)
        self.bias_initializer = tf.keras.initializers.get(bias_initializer)

    @property
    def kernel_size(self):
        return self._kernel_size

    @kernel_size.setter
    def kernel_size(self, value):
        if isinstance(value, int):
            self._kernel_size = (value, value, value)
        else:
            self._kernel_size = value

    def build(self, input_shape):
        """
        Build the layer in preparation to be trained or called. Should not be called directly,
        but rather is called when the layer is added to a model
        """
        input_shape = [
            tuple(shape.as_list()) for shape in input_shape
        ]
        if len(set(input_shape)) != 1:
            raise ValueError(
                f'All input shapes must be equal, got {input_shape}'
            )

        simplified_shape = input_shape[0]

        self.w = self.add_weight(
            shape=(len(input_shape), self.kernel_size[0], self.kernel_size[1],
                   self.kernel_size[2], simplified_shape[-1], self.filters),
            initializer=self.kernel_initializer,
            trainable=True,
            name='weights'
        )

        if self.use_bias:
            self.b = self.add_weight(
                shape=(len(input_shape), self.filters),
                initializer=self.bias_initializer,
                trainable=True,
                name='bias'
            )

    def call(self, inputs):
        """
        This is where the layer's logic lives and is called upon inputs

        Parameters
        ----------
        inputs : TensorFlow Tensor or Tensor-like
            The inputs to the layer

        Returns
        -------
        outputs : TensorFlow Tensor
            The outputs of the layer's logic
        """
        conv_outputs = [
            tf.nn.convolution(
                inputs[i],
                self.w[i],
                padding=self.padding.upper() if isinstance(
                    self.padding, str) else self.padding,
                strides=self.strides,
                data_format='NDHWC'
            ) for i in range(len(inputs))
        ]
        if self.use_bias:
            conv_outputs = [
                conv_outputs[i] + self.b[i] for i in range(len(conv_outputs))
            ]
        return [self.activation(output) for output in conv_outputs]

    def get_config(self):
        config = super().get_config().copy()
        config.update(
            {
                'filters': self.filters,
                'kernel_size': self.kernel_size,
                'padding': self.padding,
                'strides': self.strides,
                'use_bias': self.use_bias,
                'activation': tf.keras.activations.serialize(self.activation),
                'kernel_initializer': tf.keras.initializers.serialize(self.kernel_initializer),
                'bias_initializer': tf.keras.initializers.serialize(self.bias_initializer)
            }
        )
        return config

    @classmethod
    def from_config(cls, config):
        return cls(
            filters=config['filters'],
            kernel_size=config['kernel_size'],
            padding=config['padding'],
            strides=config['strides'],
            activation=config['activation'],
            use_bias=config['use_bias'],
            kernel_initializer=config['kernel_initializer'],
            bias_initializer=config['bias_initializer']
        )



import torch


class SparseMultiDense(torch.nn.Module):
    """
    Sparse implementation of the Multi-Fully-Connected layer
    """

    def __init__(
            self,
            weight,
            bias,
            device=None,
            dtype=None
    ):
        """
        Parameters
        ----------
        weight : torch.Tensor or Tensor-like
            The weight to use
        bias : torch.Tensor or Tensor-like
            The bias to use
        """

        factory_kwargs = {'device': device, 'dtype': dtype}
        super().__init__()
        for i in range(weight.shape[0]):
            self.register_buffer(
                f'w_{i}',
                torch.Tensor(weight[i]).to(**factory_kwargs).to_sparse()
            )
            self.register_buffer(
                f'b_{i}',
                torch.Tensor(bias[i]).to(**factory_kwargs).to_sparse()
            )

    def forward(self, inputs):
        """
        Call the layer on input data

        Parameters
        ----------
        inputs : torch.Tensor
            Inputs to call the layer's logic on

        Returns
        -------
        results : torch.Tensor
            The results of the layer's logic
        """
        outputs = []
        for i in range(len(inputs)):
            out = torch.sparse.mm(
                self.get_buffer(f'w_{i}').t(),
                inputs[i].t()
            ).t()
            out = torch.add(
                out,
                self.get_buffer(f'b_{i}').to_dense()
            )
            outputs.append(out)
        return outputs



from beyondml.pt.layers import MaskedConv2D, MaskedConv3D, MaskedDense, MultiMaskedConv2D, MultiMaskedConv3D, MultiMaskedDense, MultitaskNormalization


def prune_model(model, percentile):
    """
    Prune a compatible model

    Parameters
    ----------
    model : PyTorch model
        A model that has been developed to have a `.layers` property containing layers to be pruned
    percentile : int
        An integer between 0 and 99 which corresponds to how much to prune the model

    Returns
    -------
    pruned_model : PyTorch model
        The pruned model

    Notes
    -----
    - The model input **must** have a `.layers` property to be able to function. Only layers within the
      `.layers` property which are recognized as prunable are pruned, via their own `.prune()` method
    - Also acts on the model in place, but returns the model for ease of use
    """

    compatible_layers = (MaskedConv2D, MaskedConv3D, MaskedDense,
                         MultiMaskedConv2D, MultiMaskedConv3D, MultiMaskedDense)

    try:
        for layer in model.layers:
            if isinstance(layer, compatible_layers):
                layer.prune(percentile)
    except AttributeError:
        raise AttributeError('Input model does not have a `.layers` attribute. Please make sure to add that attribute\
        to the model class in order to use this function')

    return model



import torch


class SparseDense(torch.nn.Module):
    """
    Sparse implementation of a fully-connected layer
    """

    def __init__(
        self,
        weight,
        bias,
        device=None,
        dtype=None
    ):
        """
        Parameters
        ----------
        weight : torch.Tensor or Tensor-like
            The weight to use
        bias : torch.Tensor or Tensor-like
            The bias to use
        """

        factory_kwargs = {'device': device, 'dtype': dtype}
        super().__init__()
        self.register_buffer('w', torch.Tensor(
            weight).to(**factory_kwargs).to_sparse())
        self.register_buffer('b', torch.Tensor(
            bias).to(**factory_kwargs).to_sparse())

    def forward(self, inputs):
        """
        Call the layer on input data

        Parameters
        ----------
        inputs : torch.Tensor
            Inputs to call the layer's logic on

        Returns
        -------
        results : torch.Tensor
            The results of the layer's logic
        """
        out = torch.sparse.mm(self.w.t(), inputs.t()).t()
        out = torch.add(out, self.b.to_dense())
        return out



from tensorflow.keras.layers import Layer
import tensorflow as tf


class SparseMultiConv2D(Layer):
    """
    Sparse implementation of the MultiConv layer. If used in a model, must be saved and loaded via pickle
    """

    def __init__(
        self,
        filters,
        bias,
        padding='same',
        strides=1,
        activation=None,
        **kwargs
    ):
        """
        Parameters
        ----------
        filters : tf.Tensor
            The convolutional filters
        bias : tf.Tensor
            the bias tensor
        padding : str, int, or tuple of int (default 'same')
            The padding to use
        strides : int or tuple of int (default 1)
            The strides to use
        activation : None, str, or keras activation function (default None)
            The activation function to use
        """
        super().__init__(**kwargs)
        self.w = {
            i: tf.sparse.from_dense(filters[i]) for i in range(filters.shape[0])
        }
        self.b = {
            i: tf.sparse.from_dense(bias[i]) for i in range(bias.shape[0])
        }
        self.padding = padding
        self.strides = strides
        self.activation = tf.keras.activations.get(activation)

    def build(self, input_shapes):
        """
        Build the layer in preparation to be trained or called. Should not be called directly,
        but rather is called when the layer is added to a model
        """
        pass

    def call(self, inputs):
        """
        This is where the layer's logic lives and is called upon inputs

        Parameters
        ----------
        inputs : TensorFlow Tensor or Tensor-like
            The inputs to the layer

        Returns
        -------
        outputs : TensorFlow Tensor
            The outputs of the layer's logic
        """

        conv_outputs = [
            tf.nn.convolution(
                inputs[i],
                tf.sparse.to_dense(self.w[i]),
                padding=self.padding.upper() if isinstance(
                    self.padding, str) else self.padding,
                strides=self.strides,
                data_format='NHWC'
            ) for i in range(len(inputs))
        ]
        conv_outputs = [
            conv_outputs[i] + tf.sparse.to_dense(self.b[i]) for i in range(len(conv_outputs))
        ]
        return [
            self.activation(output) for output in conv_outputs
        ]

    def get_config(self):
        config = super().get_config().copy()
        config.update(
            {
                'padding': self.padding,
                'strides': self.strides,
                'activation': tf.keras.activations.serialize(self.activation)
            }
        )
        return config

    @classmethod
    def from_layer(cls, layer):
        """
        Create a layer from an instance of another layer
        """
        weights = layer.get_weights()
        w = weights[0]
        b = weights[1]
        padding = layer.padding
        strides = layer.strides
        activation = layer.activation
        return cls(
            w,
            b,
            padding,
            strides,
            activation
        )

    @classmethod
    def from_config(cls, config):
        return cls(**config)



from tensorflow.keras.layers import Layer


class SelectorLayer(Layer):
    """
    Layer which selects individual inputs

    Example:

    >>> # Create a model with two inputs and one SelectorLayer
    >>> input_1 = tf.keras.layers.Input(10)
    >>> input_2 = tf.keras.layers.Input(10)
    >>> selector = mann.layers.SelectorLayer(1)([input_1, input_2]) # 1 here indicates to select the second input and return it
    >>> model = tf.keras.models.Model([input_1, input_2], selector)
    >>> model.compile()
    >>> # Call the model
    >>> data1 = np.arange(10).reshape((1, 10))
    >>> data2 = 2*np.arange(10).reshape((1, 10))
    >>> model.predict([data1, data2])
    array([[ 0.,  2.,  4.,  6.,  8., 10., 12., 14., 16., 18.]], dtype=float32)

    """

    def __init__(
        self,
        sel_index,
        **kwargs
    ):
        """
        Parameters
        ----------
        sel_index : int
            The index of the inputs to be selected
        """
        super(SelectorLayer, self).__init__(**kwargs)
        self.sel_index = sel_index

    @property
    def sel_index(self):
        return self._sel_index

    @sel_index.setter
    def sel_index(self, value):
        if not isinstance(value, int):
            raise TypeError(
                f'sel_index must be int, got {value}, type {type(value)}')
        self._sel_index = value

    def call(self, inputs):
        """
        This is where the layer's logic lives and is called upon inputs

        Parameters
        ----------
        inputs : TensorFlow Tensor or Tensor-like
            The inputs to the layer

        Returns
        -------
        outputs : TensorFlow Tensor
            The outputs of the layer's logic
        """
        return inputs[self.sel_index]

    def get_config(self):
        config = super().get_config().copy()
        config.update(
            {
                'sel_index': self.sel_index
            }
        )
        return config

    @classmethod
    def from_config(cls, config):
        return cls(
            sel_index=config['sel_index']
        )



"""Some utilities to use when building, loading, and training MANN models"""

from .utils import get_custom_objects, mask_model, remove_layer_masks, add_layer_masks, quantize_model, get_task_masking_gradients, mask_task_weights, train_model_iteratively, train_model, ActiveSparsification
from .transformer import build_token_position_embedding_block, build_transformer_block



import numpy as np
import torch


class MultiMaskedDense(torch.nn.Module):
    """
    Multi-Fully-Connected layer which supports masking and pruning
    """

    def __init__(
        self,
        in_features,
        out_features,
        num_tasks,
        device=None,
        dtype=None
    ):
        """
        Parameters
        ----------
        in_features : int
            The number of input features
        out_features : int
            The number of output features.
            Also known as the number of artificial neurons
        num_tasks : int
            The number of tasks to initialize for
        """

        super().__init__()
        factory_kwargs = {'device': device, 'dtype': dtype}
        self.in_features = in_features
        self.out_features = out_features
        self.num_tasks = num_tasks

        weight = torch.Tensor(
            num_tasks,
            in_features,
            out_features
        ).to(**factory_kwargs)
        weight = torch.nn.init.kaiming_normal_(weight, a=np.sqrt(5))
        self.w = torch.nn.Parameter(weight)
        self.register_buffer(
            'w_mask', torch.ones_like(self.w, **factory_kwargs))

        bias = torch.zeros(num_tasks, out_features, **factory_kwargs)
        self.b = torch.nn.Parameter(bias)
        self.register_buffer('b_mask', torch.ones_like(bias, **factory_kwargs))

    def forward(self, inputs):
        """
        Call the layer on input data

        Parameters
        ----------
        inputs : torch.Tensor
            Inputs to call the layer's logic on

        Returns
        -------
        results : torch.Tensor
            The results of the layer's logic
        """
        outputs = []
        for i in range(len(inputs)):
            out = torch.mm(inputs[i], self.w[i] * self.w_mask[i])
            out = torch.add(out, self.b[i] * self.b_mask[i])
            outputs.append(out)
        return outputs

    def prune(self, percentile):
        """
        Prune the layer by updating the layer's mask

        Parameters
        ----------
        percentile : int
            Integer between 0 and 99 which represents the proportion of weights to be inactive

        Notes
        -----
        Acts on the layer in place
        """
        w_copy = np.abs(self.w.detach().cpu().numpy())
        b_copy = np.abs(self.b.detach().cpu().numpy())
        new_w_mask = np.zeros_like(w_copy)
        new_b_mask = np.zeros_like(b_copy)

        for task_num in range(self.num_tasks):
            if task_num != 0:
                for prev_idx in range(task_num):
                    w_copy[task_num][new_w_mask[prev_idx] == 1] = 0
                    b_copy[task_num][new_b_mask[prev_idx] == 1] = 0

            w_percentile = np.percentile(w_copy[task_num], percentile)
            b_percentile = np.percentile(b_copy[task_num], percentile)

            new_w_mask[task_num] = (
                w_copy[task_num] >= w_percentile).astype(int)
            new_b_mask[task_num] = (
                b_copy[task_num] >= b_percentile).astype(int)

        self.w_mask[:] = torch.Tensor(new_w_mask)
        self.b_mask[:] = torch.Tensor(new_b_mask)

        self.w = torch.nn.Parameter(
            self.w.detach() * self.w_mask
        )
        self.b = torch.nn.Parameter(
            self.b.detach() * self.b_mask
        )



from beyondml.tflow.layers import MaskedDense, MaskedConv2D, MaskedConv3D, FilterLayer, SumLayer, SelectorLayer, MultiMaskedDense, MultiMaskedConv2D, MultiMaskedConv3D, MultiDense, MultiConv2D, MultiConv3D, MultiMaxPool2D, MultiMaxPool3D, SparseDense, SparseConv2D, SparseConv3D, SparseMultiDense, SparseMultiConv2D, SparseMultiConv3D, MultitaskNormalization
import tensorflow as tf
from tensorflow.keras.callbacks import Callback
import numpy as np
import warnings

MASKING_LAYERS = (MaskedDense, MaskedConv2D, MaskedConv3D,
                  MultiMaskedDense, MultiMaskedConv2D, MultiMaskedConv3D)
MULTI_MASKING_LAYERS = (MultiMaskedDense, MultiMaskedConv2D, MultiMaskedConv3D)
NON_MASKING_LAYERS = (MultiDense, MultiConv2D, MultiConv3D)
SPARSE_LAYERS = (SparseDense, SparseConv2D, SparseConv3D,
                 SparseMultiDense, SparseMultiConv2D, SparseMultiConv3D)
CUSTOM_LAYERS = MASKING_LAYERS + NON_MASKING_LAYERS + SPARSE_LAYERS + \
    (FilterLayer, SumLayer, SelectorLayer, MultiMaxPool2D,
     MultiMaxPool3D, MultitaskNormalization)


class ActiveSparsification(Callback):
    """
    Keras-compatible callback object which enables active sparsification, allowing for increased sparsification as models
    train.
    """

    def __init__(
        self,
        performance_cutoff,
        performance_measure='auto',
        starting_sparsification=None,
        max_sparsification=99,
        sparsification_rate=1,
        sparsification_patience=10,
        stopping_delta=0.01,
        stopping_patience=5,
        restore_best_weights=True,
        verbose=1
    ):
        """
        Parameters
        ----------
        performance_cutoff : float
            The cutoff value that the performance measure must "beat" in order to iteratively sparsify
        performance_measure : str (default 'auto')
            The performance measure that is used in conjunction with `performance_cutoff`
        starting_sparsification : int or None (default None)
            The starting sparsification that the model has already been sparsified to. If `None`, then defaults to 0
        max_sparsification : int (default 99)
            The maximum sparsification allowed
        sparsification_rate : int (default 1)
            The increase in sparsification that occurs when model performance beats the performance cutoff
        sparsification_patience : int (default 10)
            The number of epochs the model is allowed to train for without beating the performance measure before stopping sparsification
        stopping_delta : float (default 0.01)
            The performance improvement that must be seen when pruning has stopped and early stopping is being considered
        stopping_patience : int (default 5)
            The number of epochs the model is allowed to train for without `stopping_delta` improvement before stopping training
        restore_best_weights : bool (default True)
            Whether to restore model best weights after training
        verbose : int or bool (default 1)
            Verbosity level for logging.

        Notes
        -----
        - If `performance_measure` is 'auto', defaults to the following measures, in order: 'val_accuracy', 'val_loss', 'accuracy', 'loss'
        - If `performance_measure` defaults to any accuracy value, then `performance_cutoff` represents the minimum value that must be
          beaten.  If `performance_measure` defaults to any loss value, then `performance_cutoff` represents the maximum value that must
          be beaten
        """
        super(Callback, self).__init__()
        self.performance_cutoff = performance_cutoff
        self.performance_measure = performance_measure
        self.starting_sparsification = starting_sparsification
        self.max_sparsification = max_sparsification
        self.sparsification_rate = sparsification_rate
        self.sparsification_patience = sparsification_patience
        self.stopping_delta = stopping_delta
        self.stopping_patience = stopping_patience
        self.restore_best_weights = restore_best_weights
        self.verbose = int(verbose)

    def on_train_begin(self, logs=None):
        self.prune_wait = 0
        self.stop_wait = 0

        self.best_weights = self.model.get_weights()
        self.best = None
        self.pruning = True
        self.sparsification = self.starting_sparsification if self.starting_sparsification is not None else 0

    def on_epoch_end(self, epoch, logs=None):
        if self.performance_measure == 'auto':
            if 'val_accuracy' in logs.keys():
                self.performance_measure = 'val_accuracy'
            elif 'val_loss' in logs.keys():
                self.performance_measure = 'val_loss'
            elif 'accuracy' in logs.keys():
                self.performance_measure = 'accuracy'
            else:
                self.performance_measure = 'loss'
            if self.verbose:
                print(f'Performance measure set to {self.performance_measure}')

        performance = logs[self.performance_measure]
        if self.best is None:
            self.best = performance

        if 'accuracy' in self.performance_measure:
            if self.pruning:
                if performance >= self.performance_cutoff:
                    self.best_weights = self.model.get_weights()
                    self.best = performance
                    if self.sparsification + self.sparsification_rate > self.max_sparsification:
                        print(
                            'Model cannot be sparsified further due to max sparsification parameter')
                        self.pruning = False
                    else:
                        self._sparsify_model(
                            self.sparsification + self.sparsification_rate)
                        self.sparsification = self.sparsification + self.sparsification_rate
                        self.prune_wait = 0
                        if self.verbose:
                            print(
                                f'Model performance reached {round(performance, 2)}, sparsifying to {self.sparsification}')
                else:
                    self.prune_wait += 1
                    if self.verbose:
                        print(
                            f'Model performance has not reached pruning threshold for {self.prune_wait} epoch(s)')
                    if self.prune_wait >= self.sparsification_patience:
                        self.pruning = False
                        self.model.set_weights(self.best_weights)
                        if self.verbose:
                            print(
                                f'Model performance has not reached pruning threshold for {self.prune_wait} epochs, reverting to {self.sparsification - self.sparsification_rate} sparsification and beginning early stopping')
            else:
                if performance >= self.best + self.stopping_delta:
                    self.best_weights = self.model.get_weights()
                    self.best = performance
                    self.stop_wait = 0
                    if self.verbose:
                        print(
                            f'Model performance improved to {round(self.best, 2)}')
                else:
                    self.stop_wait += 1
                    if self.verbose:
                        print(
                            f'Early stopping performance has not met threshold for {self.stop_wait} epochs')
                    if self.stop_wait >= self.stopping_patience:
                        if self.restore_best_weights:
                            self.model.set_weights(self.best_weights)
                        if self.verbose:
                            print(
                                'Model performance has not met early stopping criteria. Stopping training')
                        self.model.stop_training = True

        else:
            if self.pruning:
                if performance <= self.performance_cutoff:
                    self.best_weights = self.model.get_weights()
                    self.best = performance
                    if self.sparsification + self.sparsification_rate > self.max_sparsification:
                        print(
                            'Model cannot be sparsified further due to max sparsification parameter')
                        self.pruning = False
                    else:
                        self._sparsify_model(
                            self.sparsification + self.sparsification_rate)
                        self.sparsification = self.sparsification + self.sparsification_rate
                        self.prune_wait = 0
                        if self.verbose:
                            print(
                                f'Model performance reached {round(performance, 2)}, sparsifying to {self.sparsification}')
                else:
                    self.prune_wait += 1
                    if self.verbose:
                        print(
                            f'Model performance has not reached pruning threshold for {self.prune_wait} epoch(s)')
                    if self.prune_wait >= self.sparsification_patience:
                        self.pruning = False
                        self.model.set_weights(self.best_weights)
                        if self.verbose:
                            print(
                                f'Model performance has not reached pruning threshold for {self.prune_wait} epochs, reverting to {self.sparsification - self.sparsification_rate} sparsification and beginning early stopping')
            else:
                if performance <= self.best - self.stopping_delta:
                    self.best_weights = self.model.get_weights()
                    self.best = performance
                    self.stop_wait = 0
                    if self.verbose:
                        print(
                            f'Model performance improved to {round(self.best, 2)}')
                else:
                    self.stop_wait += 1
                    if self.verbose:
                        print(
                            f'Early stopping performance has not met threshold for {self.stop_wait} epochs')
                    if self.stop_wait >= self.stopping_patience:
                        if self.restore_best_weights:
                            self.model.set_weights(self.best_weights)
                        if self.verbose:
                            print(
                                'Model performance has not met early stopping criteria. Stopping training')
                        self.model.stop_training = True

    def _sparsify_model(self, percentage):
        """Function to sparsify the model"""
        try:
            new_model = tf.keras.models.Model.from_config(
                self.model.get_config(), custom_objects=get_custom_objects())
        except Exception:
            new_model = tf.keras.models.Sequential.from_config(
                self.model.get_config(), custom_objects=get_custom_objects())
        new_model.set_weights(self.model.get_weights())
        self.model.set_weights(
            mask_model(
                new_model,
                percentage,
                method='magnitude'
            ).get_weights()
        )


def _get_masking_gradients(
        model,
        x,
        y
):
    """
    Obtain masking layer gradients with respect to the tasks presented

    Parameters
    ----------
    model : tf.keras Model
        The model to get the gradients of
    x : np.array or array-like
        The input data
    y : np.array or array-like
        The true output

    Returns
    -------
    masking_gradients : list
        A list of gradients for the masking weights for the model
    """

    # Check outputs
    if isinstance(y, list):
        if not all([len(val.shape) > 1 for val in y]):
            raise ValueError(
                'Error in output shapes. If any tasks have a single output, please reshape the value using the `.reshape(-1, 1)` method')
    elif not len(y.shape) > 1:
        raise ValueError(
            'Error in output shapes. If your task has a single output, please reshape the value using the `.reshape(-1, 1)` method')

    # Grab the weights for the masking layers
    masking_weights = [
        layer.trainable_weights for layer in model.layers if isinstance(layer, MASKING_LAYERS)
    ]

    # Setup and obtain the losses
    losses = model.loss
    if not isinstance(losses, list):
        if callable(losses):
            losses = [losses] * len(x)
        losses = [tf.keras.losses.get(losses)] * len(x)
    else:
        losses = [tf.keras.losses.get(loss) if not callable(
            loss) else loss for loss in losses]

    # Grab the gradients for the specified weights
    with tf.GradientTape() as tape:
        raw_preds = model(x)
        losses = [losses[i](y[i], raw_preds[i]) for i in range(len(losses))]
        gradients = tape.gradient(losses, masking_weights)
    return gradients


def get_custom_objects():
    """Return a dictionary of custom objects (layers) to use when loading models trained using this package"""
    return dict(
        zip(
            ['MaskedDense', 'MaskedConv2D', 'MaskedConv3D', 'MultiMaskedDense', 'MultiMaskedConv2D', 'MultiMaskedConv3D', 'MultiDense',
                'MultiConv2D', 'MultiConv3D', 'SparseDense', 'SparseConv2D', 'SparseConv3D', 'SparseMultiDense', 'SparseMultiConv2D', 'SparseMultiConv3D',
                'FilterLayer', 'SumLayer', 'SelectorLayer', 'MultiMaxPool2D', 'MultiMaxPool3D'],
            CUSTOM_LAYERS
        )
    )


def mask_model(
        model,
        percentile,
        method='gradients',
        exclusive=True,
        x=None,
        y=None
):
    """
    Mask the multitask model for training respective using the gradients for the tasks at hand

    Parameters
    ----------
    model : keras model with MANN masking layers
        The model to be masked
    percentile : int
        Percentile to use in masking. Any weights less than the `percentile` value will be made zero
    method : str (default 'gradients')
        One of either 'gradients' or 'magnitude' - the method for how to identify weights to mask
        If method is 'gradients', utilizes the gradients with respect to the passed x and y variables
        to identify the subnetwork to activate for each task
        If method is 'magnitude', uses the magnitude of the weights to identify the subnetwork to activate for each task
    exclusive : bool (default True)
        Whether to restrict previously-used weight indices for each task. If `True`, this identifies disjoint subsets of
        weights within the layer which perform the tasks requested.
    x : list of np.ndarray or array-like
        The training data input values, ignored if "method" is 'magnitude'
    y : list of np.ndarray or array-like
        The training data output values, ignored if "method" is 'magnitude'
    """

    # Check method
    method = method.lower()
    if method not in ['gradients', 'magnitude']:
        raise ValueError(
            f"method must be one of 'gradients', 'magnitude', got {method}")

    # Get the gradients
    if method == 'gradients':
        grads = _get_masking_gradients(
            model,
            x,
            y
        )

        # Work to identify the right weights if exclusive
        if exclusive:
            gradient_idx = 0
            for layer in model.layers:
                if isinstance(layer, tf.keras.models.Model):
                    warnings.warn(
                        'mask_model does not effectively support models with models as layers if method is "gradients". Please set method to "magnitude"', RuntimeWarning)
                if isinstance(layer, MASKING_LAYERS):
                    if not isinstance(layer, MULTI_MASKING_LAYERS):
                        layer_grads = [np.abs(grad)
                                       for grad in grads[gradient_idx]]
                        new_masks = [(grad >= np.percentile(grad, percentile)).astype(
                            int) for grad in layer_grads]
                        layer.set_masks(new_masks)
                    else:
                        layer_grads = [np.abs(grad.numpy())
                                       for grad in grads[gradient_idx]]
                        new_masks = []
                        for grad in layer_grads:
                            new_mask = np.zeros(grad.shape)
                            used_weights = np.zeros(grad.shape[1:])
                            for task_idx in range(grad.shape[0]):
                                grad[task_idx][used_weights == 1] = 0
                                new_mask[task_idx] = (grad[task_idx] >= np.percentile(
                                    grad[task_idx], percentile)).astype(int)
                                used_weights += new_mask[task_idx]
                            new_masks.append(new_mask)
                        layer.set_masks(new_masks)
                    gradient_idx += 1
        # Work to identify the right weights if not exclusive
        else:
            gradient_idx = 0
            for layer in model.layers:
                if isinstance(layer, tf.keras.models.Model):
                    warnings.warn(
                        'mask_model does not effectively support models with models as layers if method is "gradients". Please set method to "magnitude"', RuntimeWarning)
                if isinstance(layer, MASKING_LAYERS):
                    if not isinstance(layer, MULTI_MASKING_LAYERS):
                        layer_grads = [np.abs(grad.numpy())
                                       for grad in grads[gradient_idx]]
                        new_masks = [(grad >= np.percentile(grad, percentile)).astype(
                            int) for grad in layer_grads]
                        layer.set_masks(new_masks)
                    else:
                        layer_grads = [np.abs(grad.numpy())
                                       for grad in grads[gradient_idx]]
                        new_masks = []
                        for grad in layer_grads:
                            new_mask = np.zeros(grad.shape)
                            for task_idx in range(grad.shape[0]):
                                new_mask[task_idx] = (grad[task_idx] >= np.percentile(
                                    grad[task_idx], percentile)).astype(int)
                            new_masks.append(new_mask)
                        layer.set_masks(new_masks)
                    gradient_idx += 1

    # Do this is method is "magnitude"
    elif method == 'magnitude':
        for layer in model.layers:
            if isinstance(layer, MASKING_LAYERS):
                if not isinstance(layer, MULTI_MASKING_LAYERS):
                    weights = [np.abs(weight.numpy())
                               for weight in layer.trainable_weights]
                    new_masks = [
                        (weight >= np.percentile(weight, percentile)).astype(int) for weight in weights
                    ]
                    layer.set_masks(new_masks)
                else:
                    weights = [np.abs(weight.numpy())
                               for weight in layer.trainable_weights]
                    if not exclusive:
                        new_masks = [np.zeros(weight.shape)
                                     for weight in weights]
                        for weight_idx in range(len(weights)):
                            for task_idx in range(weights[weight_idx].shape[0]):
                                new_masks[weight_idx][task_idx] = (weights[weight_idx][task_idx] >= np.percentile(
                                    weights[weight_idx][task_idx], percentile)).astype(int)
                    else:
                        new_masks = [np.zeros(weight.shape)
                                     for weight in weights]
                        for weight_idx in range(len(weights)):
                            for task_idx in range(weights[weight_idx].shape[0]):
                                exclusive_weight = weights[weight_idx][task_idx] * (
                                    1 - new_masks[weight_idx][:task_idx].sum(axis=0))
                                new_masks[weight_idx][task_idx] = (exclusive_weight >= np.percentile(
                                    weights[weight_idx][task_idx], percentile)).astype(int)
                    layer.set_masks(new_masks)
            elif isinstance(layer, tf.keras.models.Model):
                mask_model(
                    layer,
                    percentile,
                    method,
                    exclusive
                )

    # Compile the model again so the effects take place
    model.compile()
    return model


def _replace_config(config):
    """
    Replace the model config to remove masking layers
    """

    layer_mapping = {
        'MaskedConv2D': 'Conv2D',
        'MaskedConv3D': 'Conv3D',
        'MaskedDense': 'Dense',
        'MultiMaskedConv2D': 'MultiConv2D',
        'MultiMaskedConv3D': 'MultiConv3D',
        'MultiMaskedDense': 'MultiDense'
    }
    model_classes = ('Functional', 'Sequential')

    for i in range(len(config['layers'])):
        if config['layers'][i]['class_name'] in layer_mapping.keys():
            config['layers'][i]['class_name'] = layer_mapping[
                config['layers'][i]['class_name']
            ]
            del config['layers'][i]['config']['mask_initializer']
        elif config['layers'][i]['class_name'] in model_classes:
            config['layers'][i]['config'] = _replace_config(
                config['layers'][i]['config'])
    return config


def _create_masking_config(config):
    """
    Replace the model config to add masking layers
    """

    layer_mapping = {
        'Conv2D': 'MaskedConv2D',
        'Conv3D': 'MaskedConv3D',
        'Dense': 'MaskedDense',
        'MultiConv2D': 'MultiMaskedConv2D',
        'MultiConv3D': 'MultiMaskedConv3D',
        'MultiDense': 'MultiMaskedDense'
    }
    model_classes = ('Functional', 'Sequential')

    for i in range(len(config['layers'])):
        if config['layers'][i]['class_name'] in layer_mapping.keys():
            config['layers'][i]['class_name'] = layer_mapping[
                config['layers'][i]['class_name']
            ]
            config['layers'][i]['config']['mask_initializer'] = tf.keras.initializers.serialize(
                tf.keras.initializers.get('ones')
            )
        elif config['layers'][i]['class_name'] in model_classes:
            config['layers'][i]['config'] = _create_masking_config(
                config['layers'][i]['config'])
    return config


def _quantize_model_config(config, dtype='float16'):
    """
    Change the dtype of the model
    """

    model_classes = ('Functional', 'Sequential')

    new_config = config.copy()
    for i in range(len(new_config['layers'])):
        if new_config['layers'][i]['class_name'] in model_classes:
            new_config['layers'][i] = _quantize_model_config(
                new_config['layers'][i]['config'], dtype)
        else:
            new_config['layers'][i]['config']['dtype'] = dtype
    return new_config


def _replace_weights(new_model, old_model):
    """
    Replace the weights of a newly created model with the weights (sans masks) of an old model
    """

    for i in range(len(new_model.layers)):
        # Recursion in case the model contains other models
        if isinstance(new_model.layers[i], tf.keras.models.Model):
            _replace_weights(new_model.layers[i], old_model.layers[i])

        # If not masking layers, simply replace weights
        elif not isinstance(old_model.layers[i], MASKING_LAYERS):
            new_model.layers[i].set_weights(old_model.layers[i].get_weights())

        # If masking layers, replace only the required weights
        else:
            n_weights = len(new_model.layers[i].get_weights())
            new_model.layers[i].set_weights(
                old_model.layers[i].get_weights()[:n_weights])

    # Compile and return the model
    new_model.compile()
    return new_model


def _replace_masking_weights(new_model, old_model):
    """
    Replace the weights of a newly created model with the weights (adding masks) of an old model
    """

    for i in range(len(new_model.layers)):
        # Recursion in case the model contains other models
        if isinstance(new_model.layers[i], tf.keras.models.Model):
            _replace_masking_weights(new_model.layers[i], old_model.layers[i])

        # If not masking layers, simply replace weights
        elif not isinstance(new_model.layers[i], MASKING_LAYERS):
            new_model.layers[i].set_weights(old_model.layers[i].get_weights())

        # If masking layers, replace the weights and have all ones as the masks
        else:
            n_weights = len(old_model.layers[i].get_weights())
            weights = old_model.layers[i].get_weights()
            weights.extend(new_model.layers[i].get_weights()[n_weights:])
            new_model.layers[i].set_weights(weights)

    # Compile and return the model
    new_model.compile()
    return new_model


def remove_layer_masks(model, additional_custom_objects=None):
    """
    Convert a trained model from using Masking layers to using non-masking layers

    Parameters
    ----------
    model : TensorFlow Keras model
        The model to be converted
    additional_custom_objects : dict or None (default None)
        Additional custom layers to use

    Returns
    -------
    new_model : TensorFlow Keras model
        The converted model
    """

    custom_objects = get_custom_objects()
    if additional_custom_objects is not None:
        custom_objects.update(additional_custom_objects)

    # Replace the config of the model
    config = model.get_config()
    new_config = _replace_config(config)

    # Create the new model
    try:
        new_model = tf.keras.models.Model().from_config(
            new_config,
            custom_objects=custom_objects
        )
    except Exception:
        new_model = tf.keras.models.Sequential().from_config(
            new_config,
            custom_objects=custom_objects
        )

    # Replace the weights of the new model to be equivalent to the old model
    new_model = _replace_weights(new_model, model)

    # Make the new model not trainable and compile the model for good measure
    new_model.trainable = False
    new_model.compile()
    return new_model


def add_layer_masks(model, additional_custom_objects=None):
    """
    Convert a trained model from one that does not have masking weights to one that does have
    masking weights

    Parameters
    ----------
    model : TensorFlow Keras model
        The model to be converted
    additional_custom_objects : dict or None (default None)
        Additional custom layers to use

    Returns
    -------
    new_model : TensorFlow Keras model
        The converted model
    """

    custom_objects = get_custom_objects()
    if additional_custom_objects is not None:
        custom_objects.update(additional_custom_objects)

    # Replace the config of the model
    config = model.get_config()
    new_config = _create_masking_config(config)

    # Create the new model
    try:
        new_model = tf.keras.models.Model().from_config(
            new_config,
            custom_objects=custom_objects
        )
    except Exception:
        new_model = tf.keras.models.Sequential().from_config(
            new_config,
            custom_objects=custom_objects
        )

    # Replace the weights of the new model
    new_model = _replace_masking_weights(new_model, model)

    # Compile and return the model
    new_model.compile()
    return new_model


def quantize_model(model, dtype='float16', additional_custom_objects=None):
    """
    Apply model quantization

    Parameters
    ----------
    model : TensorFlow Keras Model
        The model to quantize
    dtype : str or TensorFlow datatype (default 'float16')
        The datatype to quantize to
    additional_custom_objects : None or dict (default None)
        Additional custom  objects to use to instantiate the model

    Returns
    -------
    new_model : TensorFlow Keras Model
        The quantized model
    """

    # Grab the configuration from the original model
    model_config = model.get_config()

    # Grab the weights from the original model as well
    weights = model.get_weights()

    # Change the weights to have the new datatype
    new_weights = [
        np.array(w, dtype=dtype) for w in weights
    ]

    # Change the config to get the quantized configuration
    new_config = _quantize_model_config(model_config, dtype)

    # Instantiate the new model from the new config
    custom_objects = get_custom_objects()
    if additional_custom_objects is not None:
        custom_objects.update(additional_custom_objects)
    try:
        new_model = tf.keras.models.Model.from_config(
            new_config, custom_objects=custom_objects)
    except Exception:
        new_model = tf.keras.models.Sequential.from_config(
            new_config, custom_objects=custom_objects)

    # Set the weights of the new model
    new_model.set_weights(new_weights)
    return new_model


def _get_masking_weights(model):
    """
    Get the masking weights of a model

    Parameters
    ----------
    model : TensorFlow Keras model
        The model to get the masking weights of

    Returns
    -------
    weights : list of TensorFlow tensors
        The requested weights
    """
    return [
        layer.weights for layer in model.layers if isinstance(layer, MASKING_LAYERS)
    ]


def get_task_masking_gradients(
    model,
    task_num
):
    """
    Get the gradients of masking weights within a model

    Parameters
    ----------
    model : TensorFlow Keras model
        The model to retrieve the gradients of

    Notes
    -----
    - This function should only be run *before* the model has been trained
        or used to predict.  There is an unknown bug related to TensorFlow which
        is leading to incorrect results after initial training
    - When running this function, randomized input and output data is sent
        through the model to retrieve gradients respective to each task. If
        the model is compiled using `sparse_categorical_crossentropy' loss,
        this will break this function's functionality. As a result, please
        use `categorical_crossentropy` (or even better, `mse`) before running this function. After
        retrieving gradients, the model can be recompiled with whatever parameters are desired.

    Returns
    -------
    gradients : list of TensorFlow tensors
        The gradients of the masking weights of the model
    """
    # Figure out the number of tasks
    output_shapes = model.output_shape
    if isinstance(output_shapes, list):
        num_tasks = len(output_shapes)
    else:
        num_tasks = 1

    # Get the loss weights
    if num_tasks > 1:
        loss_weights = [0] * num_tasks
        loss_weights[task_num] = 1

    # Get the masking weights
    masking_weights = _get_masking_weights(model)

    # Configure inputs
    inputs = []
    input_shapes = model.input_shape
    if isinstance(input_shapes, list):
        for shape in input_shapes:
            new_shape = list(shape)
            for i in range(len(new_shape)):
                if new_shape[i] is None:
                    new_shape[i] = 1
            inputs.append(np.random.random(new_shape))
    else:
        new_shape = list(input_shapes)
        for i in range(len(new_shape)):
            if new_shape[i] is None:
                new_shape[i] = 1
        inputs.append(np.random.random(new_shape))

    # Configure outputs
    outputs = []
    output_shapes = model.output_shape
    if isinstance(output_shapes, list):
        for shape in output_shapes:
            new_shape = list(shape)
            for i in range(len(new_shape)):
                if new_shape[i] is None:
                    new_shape[i] = 1
            outputs.append(np.random.random(new_shape))
    else:
        new_shape = list(output_shapes)
        for i in range(len(new_shape)):
            if new_shape[i] is None:
                new_shape[i] = 1
        outputs.append(np.random.random(new_shape))

    # Configure the losses
    losses = model.loss
    if not isinstance(losses, list):
        losses = [losses] * num_tasks
    losses = [
        tf.keras.losses.get(loss) for loss in losses
    ]

    # Get the gradients of the weights wrt the task
    with tf.GradientTape() as tape:
        raw_preds = model(inputs)
        loss_values = [losses[i](outputs[i], raw_preds[i]) * loss_weights[i]
                       for i in range(len(losses))]
        gradients = tape.gradient(loss_values, masking_weights)

    return gradients


def mask_task_weights(
    model,
    task_masking_gradients,
    percentile,
    respect_previous_tasks=True
):
    """
    Parameters
    ----------
    model : TensorFlow Keras model
        The model to be masked
    task_masking_gradients : list of TensorFlow tensors
        The gradients for the specific task requested
    percentile : int
        The percentile to mask/prune
    respect_previous_tasks : bool (default True)
        Whether to respect the weights used for previous tasks and not use them
        for subsequent tasks

    Returns
    -------
    masked_model : TensorFlow Keras model
        The masked model
    """

    # Get the actual weights to be able to set them
    masking_weights = [
        layer.get_weights() for layer in model.layers if isinstance(layer, MASKING_LAYERS)
    ]

    # Iterate through each of the layers of the model, keeping track of the index of which masking layer has been achieved
    masking_idx = 0
    for layer in model.layers:
        if isinstance(layer, MASKING_LAYERS):

            # Set the new masks to be masked
            new_masks = []

            # Different procedures if multi masking layer vs single masking layer
            if isinstance(layer, MULTI_MASKING_LAYERS):

                # Check for all of the weights in the list of weights (corresponding to gradients)
                for weight_num in range(len(task_masking_gradients[masking_idx])):

                    # Set the weight and gradient values so it's easier to follow
                    weight = masking_weights[masking_idx][weight_num]
                    gradient = task_masking_gradients[masking_idx][weight_num]

                    # If gradient is None, then the value is a mask
                    task_idx_num = None
                    if gradient is not None:

                        # Figure out which task index is the right one
                        for task_idx in range(gradient.shape[0]):
                            if not (gradient[task_idx].numpy() == 0).all():
                                task_idx_num = task_idx

                        if task_idx_num is not None:
                            # Get the new weight for that task only
                            task_weight = np.abs(weight[task_idx_num])

                            # Enforce respecting previous-task weights
                            if respect_previous_tasks and task_idx_num > 0:
                                task_weight[(weight[:task_idx_num] != 0).astype(
                                    int).sum(axis=0).astype(bool)] = 0

                            # Get the new mask
                            weight_mask = (task_weight >= np.percentile(
                                task_weight, percentile))

                            # Find the existing mask and set the value of only the task-specific part
                            layer_mask = masking_weights[masking_idx][weight_num + int(
                                len(masking_weights[masking_idx]) / 2)]
                            layer_mask[task_idx_num] = weight_mask

                            # Append the new mask
                            new_masks.append(layer_mask)

            # If the layer is a single masking layer
            else:
                for weight_num in range(len(task_masking_gradients[masking_idx])):

                    # Assign the weight and the gradient for this specific layer, for sanity
                    weight = masking_weights[masking_idx][weight_num]
                    gradient = task_masking_gradients[masking_idx][weight_num]

                    # If gradient is None, then the weight is a mask
                    if gradient is not None:

                        # Only proceed if the gradient exists
                        if not (gradient.numpy() == 0).all():
                            weight = np.abs(weight)
                            weight_mask = (
                                weight >= np.percentile(weight, percentile))
                            new_masks.append(weight_mask)

            # If new masks have been identified (it's possible that this did not occur), set the new masks for that layer
            if new_masks != []:
                layer.set_masks(new_masks)

            # Lastly, increase the masking index by 1
            masking_idx += 1

    # Compile the model again and return it
    model.compile()
    return model


def train_model_iteratively(
    model,
    task_gradients,
    train_x,
    train_y,
    validation_split,
    delta,
    batch_size,
    losses,
    optimizer='adam',
    metrics=None,
    starting_pruning=0,
    pruning_rate=10,
    patience=5,
    max_epochs=100
):
    """
    Train a model iteratively on each task, first obtaining
    baseline performance on each task and then iteratively
    training and pruning each task as far back as possible while
    maintaining acceptable performance on each task

    Parameters
    ----------
    model : TensorFlow Keras model
        The model to be trained
    task_gradients : list of TensorFlow tensors
        Gradients for each task, output from the `get_task_masking_gradients` function
    train_x : list of numpy arrays, TensorFlow Datasets, or other
              data types models can train with
        The input data to use to train on
    train_y : list of numpy arrays, TensorFlow Datasets, or other
              data types model can train with
        The output data to use to train on
    validation_split : float, or list of float
        The proportion of data to use for validation
    delta : float
        The tolerance between validation losses to be considered "acceptable"
        performance to continue
    batch_size : int
        The batch size to train with
    losses : str, list, or Keras loss function
        The loss or losses to use when training
    optimizer : str, list, or Keras optimizer
        The optimizer to use when training (default 'adam')
    starting_pruning : int or list of int (default 0)
        The starting pruning rate to use for each task
    pruning_rate : int or list of int (default [10, 5, 2, 1])
        The pruning rate to use
    patience : int (default 5)
        The patience for number of epochs to wait for performance to improve sufficiently
    max_epochs : int or list of int (default 100)
        The maximum number of epochs to use for training each task
    """

    # Get some information about the training procedure, including the number of tasks
    # and the gradients for each task
    num_tasks = len(train_x)

    # Keep track of the amount of the model currently used
    amount_used = 100

    # Start the training iterations
    for task_num in range(num_tasks):

        print(f'Training task {task_num}')

        # Get the starting task pruning rate for the current task
        if isinstance(starting_pruning, int):
            task_start_pruning = starting_pruning
        else:
            task_start_pruning = starting_pruning[task_num]

        # Get the current pruning rate
        if isinstance(pruning_rate, int):
            current_pruning_rate = pruning_rate
        else:
            current_pruning_rate = pruning_rate[task_num]

        # Configure the current validation split
        if isinstance(validation_split, float):
            current_validation_split = validation_split
        else:
            current_validation_split = validation_split[task_num]

        # Configure the loss weights
        loss_weights = [0] * num_tasks
        loss_weights[task_num] = 1

        # Configure the epochs
        if isinstance(max_epochs, int):
            current_epochs = max_epochs
        else:
            current_epochs = max_epochs[task_num]

        # Compile the model
        model.compile(
            loss=losses,
            optimizer=optimizer,
            loss_weights=loss_weights,
            metrics=metrics
        )

        # Train the model initially
        callback = tf.keras.callbacks.EarlyStopping(
            min_delta=delta,
            patience=patience,
            restore_best_weights=True
        )
        history = model.fit(
            train_x[task_num],
            train_y[task_num],
            batch_size=batch_size,
            epochs=current_epochs,
            validation_split=current_validation_split,
            callbacks=[callback],
            verbose=2
        )

        # Retrieve the validation loss and current best weights
        best_loss = min(history.history['val_loss'])
        best_weights = model.get_weights()

        # Training loop for the task at hand
        current_wait = 0
        if task_num == 0:
            current_prune = task_start_pruning
        else:
            current_prune = max(task_start_pruning, amount_used)
        keep_training = True
        just_started = True

        # If pruning needs to occur, do it
        if current_prune != 0:

            print(f'Pruning task to {current_prune}')

            model = mask_task_weights(
                model,
                task_gradients[task_num],
                current_prune
            )

        # keep_training indicates that training is to occur
        while keep_training:

            # First prune the model to the next pruning rate
            if current_prune + current_pruning_rate < 100:

                # Check if the training just started
                if not just_started:
                    # Increase the pruning rate
                    current_prune += current_pruning_rate
                    model = mask_task_weights(
                        model,
                        task_gradients[task_num],
                        current_prune
                    )

                    print(f'Pruning task to {current_prune}')

                else:
                    just_started = False

                # Recompile the model
                model.compile(
                    loss=losses,
                    optimizer=optimizer,
                    loss_weights=loss_weights,
                    metrics=metrics
                )

                # Train the model with the new pruning rate
                while current_wait < patience:

                    # Fit the model for a single epoch
                    history = model.fit(
                        train_x[task_num],
                        train_y[task_num],
                        batch_size=batch_size,
                        validation_split=current_validation_split,
                        verbose=2
                    )

                    # Get the new loss
                    loss = history.history['val_loss'][-1]

                    # If loss is within acceptable range, grab the best weights and
                    # reassign the best loss. Otherwise, increase current wait
                    if loss < best_loss + delta:
                        best_weights = model.get_weights()
                        best_loss = loss
                        break
                    else:
                        current_wait += 1

                # If pruning was not successful, restore best pruning rate
                if current_wait == patience or current_prune + current_pruning_rate >= 100:
                    keep_training = False

            else:
                keep_training = False

        # Record how much of the model has been used
        if task_num == 0:
            amount_used -= current_prune
        else:
            amount_used += 100 - current_prune

        # Now that current wait has been reached, restore best weights
        model.set_weights(best_weights)

        # Recompile the model
        model.compile(
            loss=losses,
            optimizer=optimizer,
            loss_weights=loss_weights,
            metrics=metrics
        )

        # Fit using the new best weights
        model.fit(
            train_x[task_num],
            train_y[task_num],
            batch_size=batch_size,
            epochs=current_epochs,
            validation_split=current_validation_split,
            callbacks=[callback],
            verbose=2
        )

    return model


def train_model(
    model,
    train_x,
    train_y,
    loss,
    metrics,
    optimizer,
    cutoff,
    batch_size=32,
    epochs=100,
    starting_sparsification=0,
    max_sparsification=99,
    sparsification_rate=5,
    sparsification_patience=10,
    stopping_patience=5
):
    model.compile(loss=loss, metrics=metrics, optimizer=optimizer)

    callback = ActiveSparsification(
        cutoff,
        starting_sparsification=starting_sparsification,
        max_sparsification=max_sparsification,
        sparsification_rate=sparsification_rate,
        sparsification_patience=sparsification_patience,
        stopping_patience=stopping_patience
    )

    model.fit(
        train_x,
        train_y,
        batch_size=batch_size,
        epochs=epochs,
        callbacks=[callback]
    )

    return model



import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import Layer


class MultiDense(Layer):
    """
    Multitask fully connected layer

    This layer implements multiple stacks of fully connected weights to account for different
    ways neurons can activate for various tasks. It is expected that to train using the RSN2 algorithm
    that MultiMaskedDense layers be used during training and then those layers be converted to this layer type.

    """

    def __init__(
        self,
        units,
        use_bias=True,
        activation=None,
        kernel_initializer='random_normal',
        bias_initializer='zeros',
        **kwargs
    ):
        """
        Parameters
        ----------
        units : int
            The number of artificial neurons to use
        use_bias : bool (default True)
            Whether to use a bias calculation in the outputs
        activation : None, str, or function (default None)
            The activation function to use on the outputs
        kernel_initializer : str or keras initialization function (default 'random_normal')
            The weight initialization function to use
        bias_initializer : str or keras initialization function (default 'zeros')
            The bias initialization function to use

        """
        super(MultiDense, self).__init__(**kwargs)
        self.units = int(units) if not isinstance(units, int) else units
        self.use_bias = use_bias
        self.activation = tf.keras.activations.get(activation)
        self.kernel_initializer = tf.keras.initializers.get(kernel_initializer)
        self.bias_initializer = tf.keras.initializers.get(bias_initializer)

    def build(self, input_shape):
        """
        Build the layer in preparation to be trained or called. Should not be called directly,
        but rather is called when the layer is added to a model
        """
        input_shape = [
            tuple(shape.as_list()) for shape in input_shape
        ]
        if len(set(input_shape)) != 1:
            raise ValueError(
                f'All input shapes must be equal, got {input_shape}')

        simplified_shape = input_shape[0]

        self.w = self.add_weight(
            shape=(len(input_shape), simplified_shape[-1], self.units),
            initializer=self.kernel_initializer,
            trainable=True,
            name='weights'
        )

        if self.use_bias:
            self.b = self.add_weight(
                shape=(len(input_shape), self.units),
                initializer=self.bias_initializer,
                trainable=True,
                name='bias'
            )

    def call(self, inputs):
        """
        This is where the layer's logic lives and is called upon inputs

        Parameters
        ----------
        inputs : TensorFlow Tensor or Tensor-like
            The inputs to the layer

        Returns
        -------
        outputs : TensorFlow Tensor
            The outputs of the layer's logic
        """
        output_tensor = [
            tf.matmul(inputs[i], self.w[i]) for i in range(len(inputs))
        ]
        if self.use_bias:
            output_tensor = [
                output_tensor[i] + (self.b[i]) for i in range(len(output_tensor))
            ]
        return [self.activation(tensor) for tensor in output_tensor]

    def get_config(self):
        config = super().get_config().copy()
        config.update(
            {
                'units': self.units,
                'use_bias': self.use_bias,
                'activation': tf.keras.activations.serialize(self.activation),
                'kernel_initializer': tf.keras.initializers.serialize(self.kernel_initializer),
                'bias_initializer': tf.keras.initializers.serialize(self.bias_initializer)
            }
        )
        return config

    @classmethod
    def from_config(cls, config):
        return cls(
            units=config['units'],
            use_bias=config['use_bias'],
            activation=config['activation'],
            kernel_initializer=config['kernel_initializer'],
            bias_initializer=config['bias_initializer']
        )



import torch


class SparseConv3D(torch.nn.Module):
    """
    Sparse 3D Convolutional layer, expected to be converted from a
    trained, pruned layer
    """

    def __init__(
        self,
        kernel,
        bias,
        padding='same',
        strides=1,
        device=None,
        dtype=None
    ):
        """
        Parameters
        ----------
        kernel : torch.Tensor or Tensor-like
            The kernel to use
        bias : torch.Tensor or Tensor-like
            The bias to use
        padding : str or int (default 'same')
            The padding to use
        strides : int or tuple (default 1)
            The strides to use
        """

        factory_kwargs = {'device': device, 'dtype': dtype}
        super().__init__()
        self.register_buffer('w', torch.Tensor(
            kernel).to(**factory_kwargs).to_sparse())
        self.register_buffer('b', torch.Tensor(
            bias).to(**factory_kwargs).to_sparse())

        self.padding = padding
        self.strides = strides

    def forward(
        self,
        inputs
    ):
        """
        Call the layer on input data

        Parameters
        ----------
        inputs : torch.Tensor
            Inputs to call the layer's logic on

        Returns
        -------
        results : torch.Tensor
            The results of the layer's logic
        """
        kernel = self.w.to_dense()
        bias = self.b.to_dense()

        return torch.nn.functional.conv3d(
            inputs,
            kernel,
            bias,
            stride=self.strides,
            padding=self.padding
        )



import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import Layer


class FilterLayer(Layer):
    """
    Layer which filters inputs based on status of `on` or `off`

    Example:

    >>> # Create a model with just a FilterLayer
    >>> input_layer = tf.keras.layers.Input(10)
    >>> filter_layer = mann.layers.FilterLayer()(input_layer)
    >>> model = tf.keras.models.Model(input_layer, filter_layer)
    >>> model.compile()
    >>> # Call the model with the layer turned on
    >>> data = np.arange(10).reshape((1, 10))
    >>> model.predict(data)
    array([[0., 1., 2., 3., 4., 5., 6., 7., 8., 9.]], dtype=float32)
    >>> # Turn off the FilterLayer and call it again
    >>> model.layers[-1].turn_off()
    >>> # Model must be recompiled after turning the layer on or off
    >>> model.compile()
    >>> model.predict(data)
    array([[0., 0., 0., 0., 0., 0., 0., 0., 0., 0.]], dtype=float32)

    """

    def __init__(
            self,
            is_on=True,
            **kwargs
    ):
        super(FilterLayer, self).__init__(**kwargs)
        self.is_on = is_on

    def call(self, inputs):
        """
        This is where the layer's logic lives and is called upon inputs

        Parameters
        ----------
        inputs : TensorFlow Tensor or Tensor-like
            The inputs to the layer

        Returns
        -------
        outputs : TensorFlow Tensor
            The outputs of the layer's logic
        """
        if self.is_on:
            return inputs
        else:
            return tf.zeros_like(inputs)

    def get_config(self):
        config = super().get_config().copy()
        config.update({'is_on': self.is_on})
        return config

    def turn_on(self):
        """Turn the layer `on` so inputs are returned unchanged as outputs"""
        self.is_on = True

    def turn_off(self):
        """Turn the layer `off` so inputs are destroyed and all-zero tensors are output"""
        self.is_on = False

    @classmethod
    def from_config(cls, config):
        return cls(**config)



from setuptools import setup
from beyondml import __version__

with open('requirements.txt', 'r') as f:
    requirements = [
        line for line in f.read().splitlines() if line != ''
    ]

setup(
    name='beyondml',
    version=__version__,
    url='https://github.com/Beyond-ML-Labs/BeyondML',
    packages=['beyondml', 'beyondml.tflow', 'beyondml.tflow.layers',
              'beyondml.tflow.utils', 'beyondml.pt', 'beyondml.pt.layers',
              'beyondml.pt.utils'],
    author='The AI Squared Team',
    author_email='mann@squared.ai',
    description='Package containing utilities for implementing RSN2/MANN',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    license='Apache 2.0',
    install_requires=requirements
)



import torch


class MultiConv3D(torch.nn.Module):
    """
    Multitask 3D Convolutional layer initialized with weights rather than with hyperparameters
    """

    def __init__(
        self,
        kernel,
        bias,
        padding='same',
        strides=1,
        device=None,
        dtype=None
    ):
        """
        Parameters
        ----------
        kernel : torch.Tensor or Tensor-like
            The kernel tensor to use
        bias : torch.Tensor or Tensor-like
            The bias tensor to use
        padding : str or int (default 'same')
            The padding to use
        strides : int or tuple (default 1)
            The strides to use
        """

        factory_kwargs = {'device': device, 'dtype': dtype}
        super().__init__()
        self.w = torch.nn.Parameter(
            torch.Tensor(kernel).to(**factory_kwargs)
        )
        self.b = torch.nn.Parameter(
            torch.Tensor(bias).to(**factory_kwargs)
        )

        self.padding = padding
        self.strides = strides

    def forward(
        self,
        inputs
    ):
        """
        Call the layer on input data

        Parameters
        ----------
        inputs : torch.Tensor
            Inputs to call the layer's logic on

        Returns
        -------
        results : torch.Tensor
            The results of the layer's logic
        """

        outputs = []
        for i in range(len(inputs)):
            outputs.append(
                torch.nn.functional.conv3d(
                    inputs[i],
                    self.w[i],
                    self.b[i],
                    stride=self.strides,
                    padding=self.padding
                )
            )
        return outputs



"""
## TensorFlow compatibility for building MANN models.

The `beyondml.tflow` package contains two subpackages, `beyondml.tflow.layers` and `beyondml.tflow.utils`, which contain
the functionality to create and train MANN layers within TensorFlow. For individuals who are
familiar with the former name of this package, `mann`, backwards compatibility can be achieved
(assuming only TensorFlow support is needed), by replacing the following line of code:

>>> import mann

with the following line:

>>> import beyondml.tflow as mann

in all existing scripts.

Within the `layers` package, there is current functionality for the the following layers:
- `beyondml.tflow.layers.FilterLayer`
- `beyondml.tflow.layers.MaskedConv2D`
- `beyondml.tflow.layers.MaskedDense`
- `beyondml.tflow.layers.MultiConv2D`
- `beyondml.tflow.layers.MultiDense`
- `beyondml.tflow.layers.MultiMaskedConv2D`
- `beyondml.tflow.layers.MultiMaskedDense`
- `beyondml.tflow.layers.MultiMaxPool2D`
- `beyondml.tflow.layers.SelectorLayer`
- `beyondml.tflow.layers.SumLayer`
- `beyondml.tflow.layers.SparseDense`
- `beyondml.tflow.layers.SparseConv`
- `beyondml.tflow.layers.SparseMultiDense`
- `beyondml.tflow.layers.SparseMultiConv`

**Note that with any of the sparse layers (such as the `SparseDense` layer), any model which
utilizes these layers will not be loadable using the traditional `load_model` functions available
in TensorFlow. Instead, the model should be saved using either joblib or pickle.**

Within the `utils` package, there are the current functions and classes:
- `ActiveSparsification`
- `build_transformer_block`
- `build_token_position_embedding_block`
- `get_custom_objects`
- `mask_model`
- `remove_layer_masks`
- `add_layer_masks`
- `quantize_model`
- `get_task_masking_gradients`
- `mask_task_weights`
- `train_model_iteratively`
"""

import beyondml.tflow.layers
import beyondml.tflow.utils



import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import Layer


class MultiMaskedConv3D(Layer):
    """
    Masked multitask 3-dimensional convoluational layer. This layer implements
    multiple stacks of the convolutional architecture and implements masking
    consistent with the BeyondML API to support developing sparse multitask models.
    """

    def __init__(
        self,
        filters,
        kernel_size=3,
        padding='same',
        strides=1,
        use_bias=True,
        activation=None,
        kernel_initializer='random_normal',
        bias_initializer='zeros',
        mask_initializer='ones',
        **kwargs
    ):
        """
        Parameters
        ----------
        filters : int
            The number of convolutional filters to apply
        kernel_size : int or tuple of ints (default 3)
            The kernel size in height and width
        padding : str (default 'same')
            Either 'same' or 'valid', the padding to use during convolution
        strides : int or tuple of ints
            Stride lenghts to use during convolution
        use_bias : bool (default True)
            Whether to use a bias calculation on the outputs
        activation : None, str, or function (default None)
            Activation function to use on the outputs
        kernel_initializer : str or keras initialization function (default 'random_normal')
            The initialization function to use for the weights
        bias_initializer : str or keras initialization function (default 'zeros')
            The initialization function to use for the bias
        mask_initializer : str or keras initialization function (default 'ones')
            The mask initialization function to use

        """
        super().__init__(**kwargs)
        self.filters = int(filters) if not isinstance(
            filters, int) else filters
        self.kernel_size = kernel_size
        self.padding = padding
        self.strides = tuple(strides) if isinstance(strides, list) else strides
        self.activation = tf.keras.activations.get(activation)
        self.use_bias = use_bias
        self.kernel_initializer = tf.keras.initializers.get(kernel_initializer)
        self.bias_initializer = tf.keras.initializers.get(bias_initializer)
        self.mask_initializer = tf.keras.initializers.get(mask_initializer)

    @property
    def kernel_size(self):
        return self._kernel_size

    @kernel_size.setter
    def kernel_size(self, value):
        if isinstance(value, int):
            self._kernel_size = (value, value, value)
        else:
            self._kernel_size = value

    def build(self, input_shape):
        """
        Build the layer in preparation to be trained or called. Should not be called directly,
        but rather is called when the layer is added to a model
        """
        input_shape = [
            tuple(shape.as_list()) for shape in input_shape
        ]
        if len(set(input_shape)) != 1:
            raise ValueError(
                f'All input shapes must be equal, got {input_shape}'
            )

        simplified_shape = input_shape[0]

        self.w = self.add_weight(
            shape=(len(input_shape), self.kernel_size[0], self.kernel_size[1],
                   self.kernel_size[2], simplified_shape[-1], self.filters),
            initializer=self.kernel_initializer,
            trainable=True,
            name='weights'
        )
        self.w_mask = self.add_weight(
            shape=self.w.shape,
            initializer=self.mask_initializer,
            trainable=False,
            name='weights_mask'
        )

        if self.use_bias:
            self.b = self.add_weight(
                shape=(len(input_shape), self.filters),
                initializer=self.bias_initializer,
                trainable=True,
                name='bias'
            )
            self.b_mask = self.add_weight(
                shape=self.b.shape,
                initializer=self.mask_initializer,
                trainable=False,
                name='bias_mask'
            )

    def call(self, inputs):
        """
        This is where the layer's logic lives and is called upon inputs

        Parameters
        ----------
        inputs : TensorFlow Tensor or Tensor-like
            The inputs to the layer

        Returns
        -------
        outputs : TensorFlow Tensor
            The outputs of the layer's logic
        """
        conv_outputs = [
            tf.nn.convolution(
                inputs[i],
                self.w[i] * self.w_mask[i],
                padding=self.padding.upper(),
                strides=self.strides,
                data_format='NDHWC'
            ) for i in range(len(inputs))
        ]
        if self.use_bias:
            conv_outputs = [
                conv_outputs[i] + (self.b[i] * self.b_mask[i]) for i in range(len(conv_outputs))
            ]
        return [self.activation(output) for output in conv_outputs]

    def get_config(self):
        config = super().get_config().copy()
        config.update(
            {
                'filters': self.filters,
                'kernel_size': list(self.kernel_size),
                'padding': self.padding,
                'strides': self.strides,
                'activation': tf.keras.activations.serialize(self.activation),
                'use_bias': self.use_bias,
                'kernel_initializer': tf.keras.initializers.serialize(self.kernel_initializer),
                'bias_initializer': tf.keras.initializers.serialize(self.bias_initializer),
                'mask_initializer': tf.keras.initializers.serialize(self.mask_initializer)
            }
        )
        return config

    def set_masks(self, new_masks):
        if not self.use_bias:
            self.set_weights(
                [self.w.numpy() * new_masks[0].astype(np.float32),
                 new_masks[0].astype(np.float32)]
            )
        else:
            self.set_weights(
                [self.w.numpy() * new_masks[0].astype(np.float32), self.b.numpy() * new_masks[1].astype(
                    np.float32), new_masks[0].astype(np.float32), new_masks[1].astype(np.float32)]
            )

    @classmethod
    def from_config(cls, config):
        return cls(
            filters=config['filters'],
            kernel_size=config['kernel_size'],
            padding=config['padding'],
            strides=config['strides'],
            activation=config['activation'],
            use_bias=config['use_bias'],
            kernel_initializer=config['kernel_initializer'],
            bias_initializer=config['bias_initializer'],
            mask_initializer=config['mask_initializer']
        )



import tensorflow as tf
from tensorflow.keras.layers import Layer


class MultiMaxPool2D(Layer):
    """
    Multitask Max Pooling Layer. This layer implements the Max Pooling algorithm
    across multiple inputs for developing multitask models

    """

    def __init__(
        self,
        pool_size=(2, 2),
        strides=(1, 1),
        padding='same',
        **kwargs
    ):
        """
        Parameters
        ----------
        pool_size : integer or tuple of 2 integers (default (2, 2))
            Window size over which to take the maximum
        strides : integer or tuple of 2 integers (default (1, 1))
            Stride values to move the pooling window after each step
        padding : str (default 'same')
            One of either 'same' or 'valid', case-insensitive. The
            padding to apply to the inputs

        """
        super(MultiMaxPool2D, self).__init__(**kwargs)
        self.pool_size = pool_size
        self.strides = strides
        self.padding = padding

    def call(self, inputs):
        """
        This is where the layer's logic lives and is called upon inputs

        Parameters
        ----------
        inputs : TensorFlow Tensor or Tensor-like
            The inputs to the layer

        Returns
        -------
        outputs : TensorFlow Tensor
            The outputs of the layer's logic
        """
        return [
            tf.nn.max_pool2d(
                inputs[i],
                self.pool_size,
                self.strides,
                self.padding.upper(),
                'NHWC'
            ) for i in range(len(inputs))
        ]

    def get_config(self):
        config = super().get_config().copy()
        config.update(
            {
                'pool_size': self.pool_size,
                'strides': self.strides,
                'padding': self.padding
            }
        )
        return config

    @classmethod
    def from_config(cls, config):
        return cls(
            pool_size=config['pool_size'],
            strides=config['strides'],
            padding=config['padding']
        )



import torch
from typing import Optional, Any, Union, Callable
import torch
from torch import Tensor
from torch.nn import Dropout, LayerNorm
from beyondml.pt.layers import MaskedDense
from torch.nn import functional as F
from .MaskedMultiHeadAttention import MaskedMultiHeadAttention


class MaskedTransformerEncoderLayer(torch.nn.Module):
    """TransformerEncoderLayer is made up of self-attn and feedforward network.
    Args:
        d_model: the number of expected features in the input (required).
        nhead: the number of heads in the multiheadattention models (required).
        dim_feedforward: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).
        activation: the activation function of the intermediate layer, can be a string
            ("relu" or "gelu") or a unary callable. Default: relu
        layer_norm_eps: the eps value in layer normalization components (default=1e-5).
        batch_first: If ``True``, then the input and output tensors are provided
            as (batch, seq, feature). Default: ``False`` (seq, batch, feature).
        norm_first: if ``True``, layer norm is done prior to attention and feedforward
            operations, respectivaly. Otherwise it's done after. Default: ``False`` (after).
    """
    __constants__ = ['batch_first', 'norm_first']

    def __init__(self,
                 d_model: int,
                 nhead: int,
                 dim_feedforward: int = 2048,
                 dropout: float = 0.1,
                 activation: Union[str, Callable[[Tensor],
                                                 Tensor]] = torch.nn.functional.relu,
                 layer_norm_eps: float = 1e-5,
                 batch_first: bool = False,
                 norm_first: bool = False,
                 device=None,
                 dtype=None
                 ) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        super(MaskedTransformerEncoderLayer, self).__init__()
        self.self_attn = MaskedMultiHeadAttention(
            d_model,
            nhead,
            dropout=dropout,
            batch_first=batch_first,
            **factory_kwargs
        )

        # Implementation of Feedforward model
        self.linear1 = MaskedDense(d_model, dim_feedforward, **factory_kwargs)
        self.dropout = Dropout(dropout)
        self.linear2 = MaskedDense(d_model, dim_feedforward, **factory_kwargs)

        self.norm_first = norm_first
        self.norm1 = LayerNorm(d_model, eps=layer_norm_eps, **factory_kwargs)
        self.norm2 = LayerNorm(d_model, eps=layer_norm_eps, **factory_kwargs)
        self.dropout1 = Dropout(dropout)
        self.dropout2 = Dropout(dropout)

    def __setstate__(self, state):
        super(MaskedTransformerEncoderLayer, self).__setstate__(state)
        if not hasattr(self, 'activation'):
            self.activation = F.relu

    def forward(self, src: Tensor):
        """Pass the input through the encoder layer.
        Args:
            src: the sequence to the encoder layer (required).
        """

        x = src

        x = self._sa_block(x)
        x = self._ff_block(x)

        return x

    # self-attention block
    def _sa_block(self, x: Tensor,
                  attn_mask: Optional[Tensor], key_padding_mask: Optional[Tensor]) -> Tensor:
        x = self.self_attn(x, x, x,
                           attn_mask=attn_mask,
                           key_padding_mask=key_padding_mask,
                           need_weights=False)[0]
        return self.dropout1(x)

    # feed forward block
    def _ff_block(self, x: Tensor) -> Tensor:
        x = self.linear2(self.dropout(self.activation(self.linear1(x))))
        return self.dropout2(x)

    def prune(self, percentile):
        self.self_attn.prune(percentile)
        self.linear1.prune(percentile)
        self.linear2.prune(percentile)
